"""
@file
@brief Class ShapeContainer
"""
import pprint
from .shape_result import ShapeResult


class ShapeContainer:
    """
    Stores all infered shapes as @see cl ShapeResult.

    Attributes:

    * `shapes`: dictionary `{ result name: ShapeResult }`
    * `names`: some dimensions are unknown and represented as
        variables, this dictionary keeps track of them
    * `names_rev`: reverse dictionary of `names`
    """

    def __init__(self):
        self.shapes = dict()
        self.names = dict()
        self.names_rev = dict()

    def __repr__(self):
        "usual"
        return f"{self.__class__.__name__}()"

    def __len__(self):
        "usual"
        return len(self.shapes)

    def __getitem__(self, key):
        "Retrieves one shape from its name."
        return self.shapes[key]

    def copy(self, deep=False):
        "Makes a copy."
        cont = ShapeContainer()
        cont.shapes = {k: v.copy(deep=deep) for k, v in self.shapes.items()}
        cont.names = self.names.copy()
        cont.names_rev = {k: v.copy() for k, v in self.names_rev.items()}
        return cont

    def update(self, key, value):
        """
        Updates one shape. Returns True if the shape was different.
        """
        if not isinstance(key, str):
            raise TypeError(  # pragma: no cover
                f"key must be a string not {type(key)!r}.")
        if not isinstance(value, ShapeResult):
            raise TypeError(  # pragma: no cover
                f"value must be a ShapeResult not {type(key)!r}.")
        if key not in self.shapes:
            self.shapes[key] = value
            return True
        r = self.shapes[key].merge(value)
        return r

    def __contains__(self, key):
        "Operator in."
        return key in self.shapes

    def __str__(self):
        """
        Displays.
        """
        rows = ["ShapeContainer({"]
        for k, v in self.shapes.items():
            rows.append(f"    {k!r}: {v!r}")
        rows.append("}, names={")
        for k, v in self.names.items():
            rows.append(f"    {k!r}: {v!r}")
        cst = self.get_all_constraints()
        if len(cst) > 0:
            rows.append("}, constraint={")
            for c, v in cst.items():
                rows.append(f"    {c!r}: {v!r}")
            rows.append("})")
        else:
            rows.append("})")

        return "\n".join(rows)

    def get_new_name(self, name, result_name, dim):
        """
        Returns a variable name when a dimension is not
        specified.
        """
        if name is not None and not isinstance(name, str):
            raise TypeError(  # pragma: no cover
                f"name must be string not {name!r}.")
        if name is None:
            name = ''
        if name == '' or name not in self.names:
            i = 0
            new_name = "%s_%d" % (name, i)
            while new_name in self.names:
                i += 1
                new_name = "%s_%d" % (name, i)
            self.names[new_name] = (name, result_name, dim)
            if name not in self.names_rev:
                self.names_rev[name] = []
            self.names_rev[name].append(new_name)
            return new_name
        val = self.names_rev[name]
        if len(val) != 1:
            raise RuntimeError(  # pragma: no cover
                f"Name {name!r} has more than one correspondance ({val!r}).")
        return val[0]

    def get_all_constraints(self):
        """
        Gathers all constraints.
        """
        cons = {}
        for _, v in self.shapes.items():
            if v.constraints is not None:
                for c in v.constraints:
                    if c.name not in cons:
                        cons[c.name] = []
                    cons[c.name].append(c)
        for _, v in cons.items():
            if len(v) > 1:
                v[0].merge(v[1:])
            del v[1:]
        return cons

    def get(self):
        """
        Returns the value of attribute `resolved_`
        (method `resolve()` must have been called first).
        """
        if not hasattr(self, 'resolved_') or self.resolved_ is None:
            raise AttributeError(  # pragma: no cover
                "Attribute 'resolved_' is missing. You must run "
                "method 'resolve()'.")
        return self.resolved_

    def resolve(self):
        """
        Resolves all constraints. It adds the attribute
        `resolved_`.
        """
        def vars_in_values(values):
            i_vals, s_vals = [], []
            for v in values:
                if isinstance(v, str):
                    s_vals.append(v)
                else:
                    i_vals.append(v)
            return set(i_vals), s_vals

        variables = {}
        for _, v in self.shapes.items():
            for sh in v.shape:
                if isinstance(sh, str):
                    variables[sh] = None

        # first step: resolves all constraint with integer
        dcsts = self.get_all_constraints()
        csts = []
        for li in dcsts.values():
            csts.extend(li)
        new_csts = []
        for cst in csts:
            if cst.name in variables and variables[cst.name] is None:
                if all(map(lambda n: isinstance(n, int), cst.values)):
                    variables[cst.name] = cst.values.copy()
                else:
                    new_csts.append(cst)
            else:
                raise RuntimeError(  # pragma: no cover
                    "Unable to find any correspondance for variable %r "
                    "in %r." % (cst.name, ", ".join(sorted(variables))))

        # second step: everything else, like a logic algorithm
        dim_names = set()
        csts = new_csts
        updates = 1
        while updates > 0 and len(new_csts) > 0:
            updates = 0
            new_csts = []
            for cst in csts:
                rvalues = variables[cst.name]
                ivalues, lvars = vars_in_values(cst.values)

                if len(lvars) > 0:
                    miss = 0
                    for lv in lvars:
                        if lv in variables and variables[lv] is not None:
                            ivalues |= variables[lv]
                        else:
                            miss += 1

                if miss == 0:
                    # simple case: only integers
                    if rvalues is None:
                        inter = ivalues
                    else:
                        inter = rvalues.intersection(ivalues)
                    if len(inter) == 0:
                        raise RuntimeError(  # pragma: no cover
                            "Resolution failed for variable %r, "
                            "current possibilities %r does not match "
                            "constraint %r." % (cst.name, rvalues, cst))
                    if rvalues is None or len(inter) < len(rvalues):
                        variables[cst.name] = inter
                        updates += 1
                    else:
                        continue
                elif len(dim_names) > 0:
                    # more complex case: variables
                    if len(cst.values) == 1 and len(lvars) == 1:
                        # exact mapping between cst.name and lvars[0]
                        a, b = cst.name, lvars[0]
                        if variables[a] is None and variables[b] is not None:
                            if variables[b].intersection(dim_names):
                                variables[a] = variables[b]
                                updates += 1
                                continue
                        elif variables[b] is None and variables[a] is not None:
                            if variables[a].intersection(dim_names):
                                variables[b] = variables[a]
                                updates += 1
                                continue

                new_csts.append(cst)
                csts = new_csts

            if len(new_csts) > 0 and updates == 0:
                # It means that a dimension needs to be left unknown.
                found = None
                for k, v in variables.items():
                    if v is None:
                        found = k
                if found is not None:
                    name = f"d{len(dim_names)}"
                    dim_names.add(name)
                    variables[found] = {name}
                    updates += 1
                else:
                    raise RuntimeError(  # pragma: no cover
                        f"Inconsistency in {self!r} with\n{variables!r}")

        # final
        results = {}
        for k, v in self.shapes.items():
            try:
                results[k] = v.resolve(variables)
            except RuntimeError as e:  # pragma: no cover
                raise RuntimeError(
                    "Unable to resolve shapes and constraints:\n%s"
                    "" % pprint.pformat(self.shapes)) from e
        self.resolved_ = results
        return self.resolved_



# pylint: disable=W0511,E1101,W1309,E0611,C0302,R0912,C0200,R1725,R0205,E0401,E1136,E1111
"""
@file
@brief Python implementation of `onnx.checker.check_model`.
"""
import os
import warnings
import numpy
from onnx import (
    TensorProto, TypeProto, ModelProto, AttributeProto, SequenceProto,
    OptionalProto)
from onnx.defs import onnx_opset_version, get_schema, OpSchema
from onnx.onnx_cpp2py_export.defs import SchemaError
from .. import get_ir_version


IR_VERSION = get_ir_version(onnx_opset_version())
ONNX_DOMAIN = ''
AI_ONNX_ML_DOMAIN = 'ai.onnx.ml'
AI_ONNX_TRAINING_DOMAIN = 'ai.onnx.ml.training'


class OnnxCheckError(RuntimeError):
    """
    Raised when a model fails check.

    :param msg: message
    :param proto: proto
    """

    def __init__(self, msg, proto):
        RuntimeError.__init__(self, msg)
        self.proto = proto


class UndefinedSchema:
    """
    Undefined schema.
    """

    def __init__(self, name, version, domain):
        self.name = name
        self.version = version
        self.domain = domain

    @property
    def deprecated_(self):
        "Returns False."
        return False

    def verify(self, node):
        "Verifies a, undefined node is consistent with ONNX language."
        if self.deprecated_:
            raise OnnxCheckError(  # pragma: no cover
                f"Operator '{self.name_}' has been deprecated since "
                f"version {self.since_version_}.",
                node)


class Schema(object):
    """
    Wrapper around a schema.
    """

    def __init__(self, schema):
        self.schema = schema

    def __getattr__(self, attr):
        if attr.endswith('_') and hasattr(self.schema, attr[:-1]):
            return getattr(self.schema, attr[:-1])
        return super(Schema, self).__getattribute__(attr)

    def num_inputs_allowed(self, n):
        "Not implemented yet."
        # return allowed_input_nums.count(n);
        return True

    def num_outputs_allowed(self, n):
        "Not implemented yet."
        # return allowed_input_nums.count(n);
        return True

    def verify(self, node):
        "Verifies a node is consistent with ONNX language."
        if self.deprecated_:
            raise OnnxCheckError(  # pragma: no cover
                f"Operator '{self.name_}' has been deprecated since "
                f"version {self.since_version_}.",
                node)

        # Check the number of inputs.
        if (len(node.input) < self.min_input_ or
                len(node.input) > self.max_input_):
            raise OnnxCheckError(  # pragma: no cover
                f"Node '{node.name}' has input size {len(node.input)} "
                f"not in range [min={self.min_input_}, "
                f"max={self.max_input_}].",
                node)

        if not self.num_inputs_allowed(len(node.input)):
            raise OnnxCheckError(  # pragma: no cover
                f"Node '{node.name}' has input size {len(node.input)} "
                f"not in allowed input sizes.",
                node)

        # Check the number of outputs.
        if (len(node.output) < self.min_output_ or
                len(node.output) > self.max_output_):
            raise OnnxCheckError(  # pragma: no cover
                f"Node '{node.name}' has output size {len(node.output)} "
                f"not in range [min={self.min_output_}, "
                f"max={self.max_output_}].",
                node)

        if not self.num_outputs_allowed(len(node.output)):
            raise OnnxCheckError(  # pragma: no cover
                f"Node '{node.name}' has output size {len(node.output)} "
                f"not in allowed output sizes.",
                node)

        # Check the values of inputs / outputs
        for in_idx in range(len(node.input)):
            if in_idx >= len(self.inputs_):
                if (not self.inputs_ and
                        OpSchema.FormalParameterOption.Variadic ==
                        self.inputs_.back().GetOption()):
                    # The last input formal parameter should be variadic.
                    break
            else:
                raise OnnxCheckError(  # pragma: no cover
                    f"Node '{node.name}' has more inputs ("
                    f"{len(node.input)} than declared {len(self.inputs_)}. "
                    f"in op definition.",
                    node)

            if (not node.input[in_idx] and
                    OpSchema.FormalParameterOption.Single ==
                    self.inputs_[in_idx].GetOption()):
                raise OnnxCheckError(  # pragma: no cover
                    f"Node '{node.name}' input[{in_idx}] is marked single but "
                    f"has an empty string in the graph.",
                    node)

        for out_idx in range(len(node.output)):
            if out_idx >= len(self.outputs_):
                if (not self.outputs_ and
                        OpSchema.FormalParameterOption.Variadic ==
                        self.outputs_.back().GetOption()):
                    # The last output formal parameter should be variadic.
                    break
            else:
                raise OnnxCheckError(  # pragma: no cover
                    f"Node '{node.name}' has more outputs ("
                    f"{len(node.output)} than declared {len(self.outputs_)}. "
                    f"in op definition.",
                    node)

            if (not node.output[out_idx] and
                    OpSchema.FormalParameterOption.Single ==
                    self.outputs_[out_idx].GetOption()):
                raise OnnxCheckError(  # pragma: no cover
                    f"Node '{node.name}' output[{out_idx}] is marked single but "
                    f"has an empty string in the graph.",
                    node)

        # An internal symbol is defined as starting with two underscores. Attributes
        # with names meeting this condition are considered implementation details
        # and should be ignored for the purpose of schema checking.
        def isInternalSymbol(sym):
            return len(sym) >= 2 and sym[0] == '_' and sym[1] == '_'

        # Check attributes
        seen_attr_names = set()
        for attr_proto in node.attribute:
            name = attr_proto.name

            if name in seen_attr_names:
                raise OnnxCheckError(  # pragma: no cover
                    f"Attribute '{name}' appeared multiple times.",
                    node)
            seen_attr_names.add(name)

            if name in self.attributes_:
                search = self.attributes_.index(name)
            else:
                search = -1
            expected_type = None
            if search != -1:
                expected_type = self.attributes_[search]
            elif self.allows_unchecked_attributes_ or isInternalSymbol(name):
                continue
            else:
                raise OnnxCheckError(  # pragma: no cover
                    f"Unrecognized attribute '{name}' for operator "
                    f"'{node.op_type}'.", node)

            # Type would be UNDEFINED if not set
            if attr_proto.type != expected_type:
                raise OnnxCheckError(  # pragma: no cover
                    f"Mismatched attribute type in '{node.name}' and "
                    f"attribute '{name}'.", node)

            # ref_attr_name is only valid when non-empty
            # we simply read default value if not present
            if not attr_proto.ref_attr_name:
                continue

            # if attr_proto.type != UNDEFINED
            # we consider primitive types to be set even
            # if proto3 did not output default values into the stream
            # in which case we will read the default
            if expected_type in (AttributeProto.FLOAT,
                                 AttributeProto.INT,
                                 AttributeProto.STRING):
                pass
            elif expected_type == AttributeProto.TENSOR:
                if attr_proto.t.ByteSize == 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{name}' is expected to have field "
                        f"'t'.", node)
            elif expected_type == AttributeProto.SPARSE_TENSOR:
                if attr_proto.sparse_tensor.ByteSize == 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{name}' is expected to have field "
                        f"'sparse_tensor'.", node)
            elif expected_type == AttributeProto.GRAPH:
                if attr_proto.g.ByteSize == 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{name}' is expected to have field "
                        f"'g'.", node)
                if node.op_type == 'If' and len(attr_proto.g.input) > 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{attr_proto.name}' of "
                        f"operator If with name '{node.name}' must not have "
                        f"inputs.", node)
            elif expected_type == AttributeProto.TYPE_PROTO:
                if attr_proto.tp.ByteSize == 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{name}' is expected to have field "
                        f"'tp'.", node)
            elif expected_type == AttributeProto.FLOATS:
                if attr_proto.floats.ByteSize == 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{name}' is expected to have field "
                        f"'floats'.", node)
            elif expected_type == AttributeProto.INTS:
                if attr_proto.ints.ByteSize == 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{name}' is expected to have field "
                        f"'ints'.", node)
            elif expected_type == AttributeProto.STRINGS:
                if attr_proto.strings.ByteSize == 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{name}' is expected to have field "
                        f"'strings'.", node)
            elif expected_type == AttributeProto.TENSORS:
                if attr_proto.tensors.ByteSize == 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{name}' is expected to have field "
                        f"'tensors'.", node)
            elif expected_type == AttributeProto.SPARSE_TENSORS:
                # Not adding check ... we should likely delete the check in all other
                # cases, which will not allow us to have an empty list as a valid value
                # for an attribute and this seems undesirable.
                pass
            elif expected_type == AttributeProto.GRAPHS:
                if attr_proto.graphs.ByteSize == 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{name}' is expected to have field "
                        f"'graphs'.", node)
            elif expected_type == AttributeProto.TYPE_PROTOS:
                if attr_proto.type_protos.ByteSize == 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Attribute '{name}' is expected to have field "
                        f"'type_protos'.", node)
            else:
                raise OnnxCheckError(  # pragma: no cover
                    f"Attribute '{name}' has unknown expected type.",
                    node)

        for attr in self.attributes_:
            if not attr.required:
                continue
            if attr.name not in seen_attr_names:
                raise OnnxCheckError(  # pragma: no cover
                    f"Required attribute '{attr.name}' is missing.",
                    node)


class CheckerContextDefaultRegistry:
    """
    Registry.
    """

    def get_schema(self, op_type, version, domain):
        "Accessor."
        try:
            return Schema(get_schema(op_type, version, domain))
        except SchemaError:
            return UndefinedSchema(op_type, version, domain)

    def GetSchema(self, op_type, version, domain):
        "Accessor."
        return self.get_schema(op_type, version, domain)


class CheckerContext:
    """
    Class hosting information about a graph.
    """

    def __init__(self, ctx=None):
        if ctx is None:
            self.ir_version_ = -1
            self.opset_imports_ = {}
            self.schema_registry_ = CheckerContextDefaultRegistry()
            self.model_dir_ = None
            self.is_main_graph_ = True
        else:
            self.ir_version_ = ctx.ir_version_
            self.opset_imports_ = ctx.opset_imports_.copy()
            self.schema_registry_ = ctx.schema_registry_
            self.model_dir_ = ctx.model_dir_
            self.is_main_graph_ = ctx.is_main_graph_

    def get_ir_version(self):
        "Accessor."
        return self.ir_version_

    def set_ir_version(self, v):
        "Accessor."
        self.ir_version_ = v

    def get_opset_imports(self):
        "Accessor."
        return self.opset_imports_

    def set_opset_imports(self, imps):
        "Accessor."
        self.opset_imports_ = imps

    def is_main_graph(self):
        "Accessor."
        return self.is_main_graph_

    def set_is_main_graph(self, is_main_graph):
        "Accessor."
        self.is_main_graph_ = is_main_graph

    def set_schema_registry(self, schema_registry):
        "Accessor."
        self.schema_registry_ = schema_registry

    def get_schema_registry(self):
        "Accessor."
        return self.schema_registry_

    def set_model_dir(self, model_dir):
        "Accessor."
        self.model_dir_ = model_dir

    def get_model_dir(self):
        "Accessor."
        return self.model_dir_


class LexicalScopeContext:
    """
    Construct an instance with the lexical scope from the parent graph to allow
    lookup of names from that scope via this_or_ancestor_graph_has.
    The caller must ensure parent_context remains valid for the entire lifetime
    of the new instance. Alternatively, if that cannot be guaranteed, create an
    instance with the default constructor and populate output_names with the
    values from the parent scope so the values are copied instead.
    """

    def __init__(self, parent_context=None):
        if parent_context is None:
            self.parent_context_ = None
        else:
            self.parent_context_ = parent_context.copy()
        self.output_names = set()

    def add(self, name):
        "Adds a name to the context."
        self.output_names.add(name)

    def this_graph_has(self, name):
        "Checks the context includes a specific name."
        return name in self.output_names

    def this_or_ancestor_graph_has(self, name):
        "Checks the context and its ancestor includes a specific name."
        return self.this_graph_has(name) or (
            self.parent_context_ and
            self.parent_context_.this_or_ancestor_graph_has(name))

    def copy(self):
        "Copies the instance."
        ctx = LexicalScopeContext(self.parent_context_)
        ctx.output_names = set(self.output_names)
        return ctx


def _enforce_has_field(proto, field):
    if not hasattr(proto, field):
        raise OnnxCheckError(  # pragma: no cover
            f"Field '{field}' of '{proto}' is required but missing.", proto)


def _enforce_has_repeated_field(proto, field):
    if not getattr(proto, field + '_size')():
        raise OnnxCheckError(  # pragma: no cover
            f"Repeated field '{field}' of '{proto}' is required but missing.", proto)


def _enforce_non_empty_field(proto, field):
    if not getattr(proto, field):
        raise OnnxCheckError(  # pragma: no cover
            f"Field '{field}' of '{proto}' is required to be non-empty.", proto)


def _check_value_info(value_info, ctx):
    _enforce_non_empty_field(value_info, "name")
    # Relax constraint for subgraph input/output.
    if not ctx.is_main_graph():
        return
    _enforce_has_field(value_info, "type")
    value_case = None
    for n in dir(value_info.type):
        if n.endswith('_type'):
            tt = getattr(value_info.type, n)
            if tt.ByteSize() > 0:
                if value_case is not None:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Value_info {value_info} has multiple types.",
                        value_info)
                value_case = n
    if value_case == "tensor_type":
        _enforce_has_field(tt, "elem_type")
        _enforce_has_field(tt, "shape")
    elif value_case == TypeProto.kOptionalType:
        tt = value_info.type.optional_type
        _enforce_has_field(tt, "elem_type")
    elif value_case == TypeProto.kSequenceType:
        tt = value_info.type.sequence_type
        _enforce_has_field(tt, "elem_type")
    elif value_case == TypeProto.kMapType:
        tt = value_info.type.map_type
        _enforce_has_field(tt, "key_type")
        _enforce_has_field(tt, "value_type")
    elif value_case == TypeProto.kOpaqueType:
        pass
    elif value_case == TypeProto.kSparseTensorType:
        tt = value_info.type.sparse_tensor_type
        _enforce_has_field(tt, "elem_type")
        _enforce_has_field(tt, "shape")
    else:
        raise OnnxCheckError(  # pragma: no cover
            f"Unrecognized type value case (value_info name '{value_info.name}' "
            f"value_case={value_case}.", value_info)


def _check_data_field(tensor, field, num_value_fields):
    at = getattr(tensor, field)
    has = len(at)
    if has:
        num_value_fields[0] += 1  # pylint: disable=E1137
        value_field = getattr(tensor, field)
        return value_field
    return None


def _check_field(tensor, field, value_field, nelem):
    if nelem != 0 and len(getattr(tensor, field)):
        raise OnnxCheckError(  # pragma: no cover
            f"values of data_type '{tensor.data_type} "
            f"should be stored in field '{field}' "
            f"instead of '{value_field}'.",
            tensor)


def _check_tensor(tensor, ctx):

    _enforce_has_field(tensor, "data_type")
    if tensor.data_type == TensorProto.UNDEFINED:
        raise OnnxCheckError(  # pragma: no cover
            f"Setting data_type field (tensor name '{tensor.name}' "
            f"to UNDEFINED is not allowed.", tensor)

    num_value_fields = [0]

    value_field = (
        _check_data_field(tensor, "float_data", num_value_fields) or
        _check_data_field(tensor, "int32_data", num_value_fields) or
        _check_data_field(tensor, "string_data", num_value_fields) or
        _check_data_field(tensor, "int64_data", num_value_fields) or
        _check_data_field(tensor, "raw_data", num_value_fields) or
        _check_data_field(tensor, "double_data", num_value_fields) or
        _check_data_field(tensor, "uint64_data", num_value_fields))

    num_value_fields = num_value_fields[0]

    stored_externally = (
        hasattr(tensor, 'data_location') and
        tensor.data_location == TensorProto.EXTERNAL)
    if stored_externally:
        if num_value_fields != 0:
            raise OnnxCheckError(  # pragma: no cover
                f"Data of TensorProto ( tensor name: f{tensor.name}) "
                f"is stored externally and should not have data field: "
                f"{value_field}.", tensor)

        has_location = False
        for entry in tensor.external_data():
            # if entry.has_key() and entry.has_value() and entry.key() == "location":
            if entry.has_value() and entry.key() == "location":
                has_location = True
                data_path = os.path.join(ctx.get_model_dir(), entry.value())
                # use stat to check whether the file exists
                if os.stat(data_path).st_size != 0:
                    raise OnnxCheckError(  # pragma: no cover
                        f"Data of TensorProto ( tensor name: {tensor.name} "
                        f"should be stored in {data_path}, but it doesn't "
                        "exist or is not accessible.", tensor)
        if not has_location:
            raise OnnxCheckError(  # pragma: no cover
                f"TensorProto tensor name {tensor.name} is stored externally "
                f"but doesn't have a location.",
                tensor)
        return

    nelem = 1
    for x in tensor.dims:
        nelem *= x

    if nelem == 0 and num_value_fields != 0:
        raise OnnxCheckError(  # pragma: no cover
            f"TensorProto (tensor name f{tensor.name} "
            f"is 0-element but contains data!",
            tensor)
    if nelem != 0 and num_value_fields != 1:
        raise OnnxCheckError(  # pragma: no cover
            f"TensorProto (tensor name: {tensor.name} "
            f"should contain one and only one value field.",
            tensor)
    if hasattr(tensor, 'raw_data') and len(tensor.raw_data) > 0:
        if tensor.data_type == TensorProto.STRING:
            raise OnnxCheckError(  # pragma: no cover
                f"STRING data (tensor name: f{tensor.name} "
                f"should not be stored in raw_data field",
                tensor)
    else:
        if tensor.data_type in (TensorProto.FLOAT,
                                TensorProto.COMPLEX64):
            _check_field(tensor, "float_data", value_field, nelem)
        elif tensor.data_type in (TensorProto.DOUBLE,
                                  TensorProto.COMPLEX128):
            _check_field(tensor, "double_data", value_field, nelem)
        elif tensor.data_type in (TensorProto.INT32,
                                  TensorProto.UINT8,
                                  TensorProto.INT8,
                                  TensorProto.UINT16,
                                  TensorProto.INT16,
                                  TensorProto.BOOL,
                                  TensorProto.FLOAT16,
                                  TensorProto.BFLOAT16):
            _check_field(tensor, "int32_data", value_field, nelem)
        elif tensor.data_type == TensorProto.INT64:
            _check_field(tensor, "int64_data", value_field, nelem)
        elif tensor.data_type == TensorProto.INT64:
            _check_field(tensor, "int64_data", value_field, nelem)
        elif tensor.data_type in (TensorProto.UINT32,
                                  TensorProto.UINT64):
            _check_field(tensor, "uint64_data", value_field, nelem)
        elif tensor.data_type == TensorProto.STRING:
            _check_field(tensor, "string_data", value_field, nelem)
        else:
            raise OnnxCheckError(  # pragma: no cover
                f"Unrecognized data_type (tensor name: {tensor.name} "
                f"): {tensor.data_type}.",
                tensor)


def _check_sequence(sequence, ctx):
    _enforce_has_field(sequence, "elem_type")
    if sequence.elem_type == SequenceProto.TENSOR:
        for tensor in sequence.tensor_values():
            _check_tensor(tensor, ctx)
    elif sequence.elem_type == SequenceProto.SPARSE_TENSOR:
        for sparse_tensor in sequence.sparse_tensor_values():
            _check_sparse_tensor(sparse_tensor, ctx)
    elif sequence.elem_type == SequenceProto.SEQUENCE:
        for seq in sequence.sequence_values():
            _check_sequence(seq, ctx)
    elif sequence.elem_type == SequenceProto.MAP:
        for map in sequence.map_values():
            _check_map(map, ctx)
    else:
        raise OnnxCheckError(  # pragma: no cover
            f"Sequence ( Structure name: {sequence.name}, "
            f"elem_type: {sequence.elem_type}) is not have "
            f"a valid element type.",
            sequence)


def _check_optional(optional, ctx):
    _enforce_has_field(optional, "elem_type")
    if optional.elem_type == OptionalProto.UNDEFINED:
        return
    elif optional.elem_type == OptionalProto.TENSOR:
        if optional.has_tensor_value():
            _check_tensor(optional.tensor_value(), ctx)
    elif optional.elem_type == OptionalProto.SPARSE_TENSOR:
        if optional.has_sparse_tensor_value():
            _check_sparse_tensor(optional.sparse_tensor_value(), ctx)
    elif optional.elem_type == OptionalProto.SEQUENCE:
        if optional.has_sequence_value():
            _check_sequence(optional.sequence_value(), ctx)
    elif optional.elem_type == OptionalProto.MAP:
        if (optional.has_map_value()):
            _check_map(optional.map_value(), ctx)
    else:
        raise OnnxCheckError(  # pragma: no cover
            f"Optional ( Structure name: {optional.name}, "
            f"elem_type: {optional.elem_type}) is not "
            f"have a valid element type.",
            optional)


def _check_map(map, ctx):
    _enforce_has_field(map, 'key_type')
    if map.key_type() == TensorProto.UNDEFINED:
        raise OnnxCheckError(  # pragma: no cover
            f"Setting key_type field (map name: '{map.name}') "
            f"to UNDEFINED is not allowed.",
            map)
    # Check if key is a valid type, specifically INT8, INT16, INT32, INT64,
    # UINT8, UINT16, UINT32, UINT64, or STRING.
    if map.key_type() in (TensorProto.FLOAT, TensorProto.BOOL,
                          TensorProto.FLOAT16, TensorProto.COMPLEX64,
                          TensorProto.COMPLEX128):
        raise OnnxCheckError(  # pragma: no cover
            f"Setting key_type field (map name: {map.name}) "
            f" to invalid TensorProto key_type {map.key_type()} "
            f"is not allowed",
            map)
    # MapProto will use either keys or string_keys, so only one should be > 0.
    if map.keys_size() > 0 and map.string_keys_size() > 0:
        raise OnnxCheckError(  # pragma: no cover
            f"Map (name: '{map.name}') should not "
            f"contain more than one keys field.",
            map)

    num_keys = map.keys_size() + map.string_keys_size()
    num_values = 0

    _enforce_has_field(map, 'values')
    _check_sequence(map.values(), ctx)

    if map.values().elem_type == SequenceProto.TENSOR:
        num_values = map.values().tensor_values_size()
    elif map.values().elem_type == SequenceProto.SPARSE_TENSOR:
        num_values = map.values().sparse_tensor_values_size()
    elif map.values().elem_type == SequenceProto.SEQUENCE:
        num_values = map.values().sequence_values_size()
    elif map.values().elem_type == SequenceProto.MAP:
        num_values = map.values().map_values_size()

    if num_keys != num_values:
        raise OnnxCheckError(  # pragma: no cover
            f"Length of map keys and map values are not the same "
            f"(map name: '{map.name}').",
            map)


def _parse_data(dtype, indices):
    if dtype != indices.dtype:
        raise OnnxCheckError(  # pragma: no cover
            f"Wrong element type {indices.dtype}, expected is {dtype}.",
            None)


def _check_sparse_tensor_indices_1(indices, sparse_tensor_proto, nnz):
    """
    Check that the index data stored in a SparseTensorProto is valid.
    indices: a 1-dimensional tensor; indices[i] represents the
    linearized index value for the i-th nonzero value.
    """
    dense_rank = sparse_tensor_proto.dims_size()
    dense_size = 1
    for i in range(dense_rank):
        dense_size *= sparse_tensor_proto.dims(i)
        if indices.dims(0) != nnz:
            raise OnnxCheckError(  # pragma: no cover
                f"Sparse tensor indices '{indices.name}' has "
                f"{indices.dims(0)} values, but NNZ is {nnz}.",
                sparse_tensor_proto)

    # Check if indices appear in ascending order, and if they have valid
    # values. The i-th value in index_data is the linear index of the i-th
    # non-zero value.
    index_data = _parse_data(numpy.int64, indices)

    prev_index = -1
    for i in range(nnz):
        curr_index = index_data[i]  # linearized index of i-th value
        if curr_index < 0 or curr_index >= dense_size:
            raise OnnxCheckError(  # pragma: no cover
                f"Sparse tensor '{indices.name}' index value at "
                f"position [{i}] out of range [0, {dense_size - 1}].",
                sparse_tensor_proto)
        if curr_index <= prev_index:
            raise OnnxCheckError(  # pragma: no cover
                f"Sparse tensor '{indices.name}' index value at "
                f"position [{i}] not in sorted order.",
                sparse_tensor_proto)
        prev_index = curr_index


def _check_sparse_tensor_indices_2(indices, sparse_tensor_proto, nnz):
    """
    Check that the index data stored in a SparseTensorProto is valid.
    indices: a 2-dimensional tensor; indices[i,j] represents the j-th
    index value for the i-th nonzero value.
    """
    dense_rank = sparse_tensor_proto.dims_size()
    if indices.dims(0) != nnz:
        raise OnnxCheckError(  # pragma: no cover
            f"Sparse tensor indices '{indices.name}' "
            f"first dimension size does not equal NNZ={nnz}.",
            sparse_tensor_proto)

    if indices.dims(1) != dense_rank:
        raise OnnxCheckError(  # pragma: no cover
            f"Sparse tensor indices '{indices.name}' "
            f"second dimension size does not equal "
            f"dense_rank={dense_rank}.",
            sparse_tensor_proto)

    # Check if indices appear in ascending order, and if they have valid
    # values.
    index_data = _parse_data(numpy.int64, indices)
    prev_index = -1
    for i in range(nnz):
        curr_index = 0  # linearized index of i-th value
        for j in range(dense_rank):
            index_ij = index_data[i * dense_rank + j]
            if index_ij < 0 or index_ij >= sparse_tensor_proto.dims(j):
                raise OnnxCheckError(  # pragma: no cover
                    f"Sparse tensor '{indices.name}' index value "
                    f"at position [{i}, {j}] out of range.",
                    sparse_tensor_proto)
            curr_index = curr_index * sparse_tensor_proto.dims(j) + index_ij
        if curr_index <= prev_index:
            raise OnnxCheckError(  # pragma: no cover
                f"Sparse tensor '{indices.name}' index value "
                f"at position [{i}] not in lexicographic sorted "
                "order.", sparse_tensor_proto)
        prev_index = curr_index


def _check_sparse_tensor(sparse_tensor_proto, ctx):
    _enforce_has_field(sparse_tensor_proto, "values")

    values = sparse_tensor_proto.values()
    _check_tensor(values, ctx)

    # values must be a tensor of shape [NNZ]
    # Currently we restrict the value associated with a particular index-tuple
    # to be a single value. In the future, if there is a requirement,
    # we may extend this to permit the value to be a "sub-tensor", in which
    # case values will have dimension > 1.
    if values.dims_size() != 1:
        raise OnnxCheckError(  # pragma: no cover
            f"Sparse tensor values '{values.name}' must have rank 1.",
            sparse_tensor_proto)

    nnz = values.dims(0)
    dense_rank = sparse_tensor_proto.dims_size()
    if dense_rank == 0:
        raise OnnxCheckError(  # pragma: no cover
            f"Sparse tensor '{values.name}' must have a "
            f"dense-rank > 0.", sparse_tensor_proto)

    for i in range(dense_rank):
        if sparse_tensor_proto.dims(i) <= 0:
            raise OnnxCheckError(  # pragma: no cover
                f"Sparse tensor '{values.name} dimensions "
                f"are not positive.", sparse_tensor_proto)

    if sparse_tensor_proto.has_indices():
        indices = sparse_tensor_proto.indices()
        _check_tensor(indices, ctx)
        if indices.data_type != TensorProto.INT64:
            raise OnnxCheckError(  # pragma: no cover
                f"Sparse tensor indices '{indices.name}' must have INT64 type.",
                sparse_tensor_proto)

        if indices.dims().size() == 1:
            # Indices in linearized format
            _check_sparse_tensor_indices_1(indices, sparse_tensor_proto, nnz)
            return
        if indices.dims().size() == 2:
            # Check COO-style index. E.g., an index for a 3D tensor is a 3-tuple.
            _check_sparse_tensor_indices_2(indices, sparse_tensor_proto, nnz)
            return
        raise OnnxCheckError(  # pragma: no cover
            f"Sparse tensor indices '{indices.name}' must have rank 1 or 2.",
            sparse_tensor_proto)
    elif nnz != 0:
        raise OnnxCheckError(  # pragma: no cover
            f"Sparse tensor '{values.name}' has no index values.",
            sparse_tensor_proto)


def check_attribute(attr, ctx, lex_ctx):
    """
    NB: This is a generic "attribute well-formedness" check, it doesn't
    actually test if an attribute is valid per a schema.
    """
    _enforce_non_empty_field(attr, "name")

    if ctx.get_ir_version() >= 0x00000002:
        _enforce_has_field(attr, "type")

    used_fields = 0

    def check_type(expected_type):
        if hasattr(attr, 'type') and attr.type != expected_type:
            raise OnnxCheckError(  # pragma: no cover
                f"Type field and data field mismatch in attribute '{attr.name}'.",
                attr)

    def check_singular_field(field, itype):
        if hasattr(attr, field):
            check_type(itype)
            return 1
        return 0

    def check_repeated_field(field, type):
        if getattr(attr, field + '_size')() > 0:
            check_type(type)
            return 1
        return 0

    used_fields += check_singular_field("f", AttributeProto.FLOAT)
    used_fields += check_singular_field("i", AttributeProto.INT)
    used_fields += check_singular_field("s", AttributeProto.STRING)
    used_fields += check_singular_field("t", AttributeProto.TENSOR)
    used_fields += check_singular_field("g", AttributeProto.GRAPH)
    used_fields += check_singular_field("tp", AttributeProto.TYPE_PROTO)
    used_fields += check_singular_field("sparse_tensor",
                                        AttributeProto.SPARSE_TENSOR)
    used_fields += check_repeated_field("floats", AttributeProto.FLOATS)
    used_fields += check_repeated_field("ints", AttributeProto.INTS)
    used_fields += check_repeated_field("strings", AttributeProto.STRINGS)
    used_fields += check_repeated_field("tensors", AttributeProto.TENSORS)
    used_fields += check_repeated_field("graphs", AttributeProto.GRAPHS)
    used_fields += check_repeated_field("sparse_tensors",
                                        AttributeProto.SPARSE_TENSORS)
    used_fields += check_repeated_field("type_protos",
                                        AttributeProto.TYPE_PROTOS)

    # Normally, used_fields is expected to be 1.
    # In proto3, when the value to be set is type default value
    # (say 0 for int), used_fields may be 0.
    if used_fields > 1:
        raise OnnxCheckError(  # pragma: no cover
            f"Attribute (name: '{attr.name}') should not "
            f"contain more than one value field.",
            attr)

    if not ctx.is_main_graph():
        # It's an attribute of a node in function body.
        if attr.has_ref_attr_name() and used_fields != 0:
            # The attribute proto is supposed to refer to data outside and does not
            # have its own value field set.
            raise OnnxCheckError(  # pragma: no cover
                f"Attribute (name: '{attr.name}') should refer "
                f"to attribute in parent node.",
                attr)

    if attr.has_t():
        _check_tensor(attr.t(), ctx)

    if attr.has_sparse_tensor():
        _check_sparse_tensor(attr.sparse_tensor(), ctx)

    if attr.has_g():
        subgraph_ctx = CheckerContext(ctx)
        subgraph_ctx.set_is_main_graph(False)
        _check_graph(attr.g(), subgraph_ctx, lex_ctx)

    for tensor in attr.tensors():
        _check_tensor(tensor, ctx)

    for sparse_tensor in attr.sparse_tensors():
        _check_sparse_tensor(sparse_tensor, ctx)

    if attr.graphs().size() > 0:
        subgraph_ctx = CheckerContext(ctx)
        subgraph_ctx.set_is_main_graph(False)
        for graph in attr.graphs():
            _check_graph(graph, subgraph_ctx, lex_ctx)


def _check_node(node, ctx, lex_ctx):
    _enforce_non_empty_field(node, "op_type")

    if not node.input and not node.output:
        raise OnnxCheckError(  # pragma: no cover
            f"NodeProto (name: '{node.name}', type: '{node.op_type}') "
            f"has zero input and zero output.",
            node)

    # If encounter experimental op, stop checking
    if check_is_experimental_op(node.op_type):
        warnings.warn(
            f"Warning: Checker does not support models "
            f"with experimental ops: '{node.op_type}'.")
        return

    # Resolve domain for node
    opset_imports = ctx.get_opset_imports()
    if node.domain not in opset_imports:
        raise OnnxCheckError(  # pragma: no cover
            f"No opset import for domain '{node.domain}'.",
            node)
    domain_version = opset_imports[node.domain]

    for attr in node.attribute:
        check_attribute(attr, ctx, lex_ctx)

    schema = ctx.get_schema_registry().GetSchema(
        node.op_type, domain_version, node.domain)
    if not schema:
        if node.domain in (ONNX_DOMAIN, AI_ONNX_ML_DOMAIN,
                           "ai.onnx", AI_ONNX_TRAINING_DOMAIN):
            # fail the checker if op in built-in domains has no schema
            raise OnnxCheckError(  # pragma: no cover
                f"No Op registered for '{node.op_type}' with domain_version "
                f"of {domain_version}.",
                node)
        else:
            # TODO: expose the registration of the op schemas appropriately in
            # python, so we can load and register operators in other domains
            # before we complete the above todo, let's skip the schema check for now
            pass
    elif schema.deprecated_:
        raise OnnxCheckError(  # pragma: no cover
            f"Op registered for '{node.op_type}' is deprecated "
            f"in domain_version of {domain_version}.",
            node)
    else:
        schema.verify(node)


def _check_graph(graph, ctx, parent_lex):
    _enforce_non_empty_field(graph, "name")

    for value_info in graph.input:
        _check_value_info(value_info, ctx)
    for value_info in graph.output:
        _check_value_info(value_info, ctx)

    # Inherit values available in outer scope
    # Note that we do not allow shadowing, so the presence of an already-defined
    # name is always an error.
    lex_ctx = LexicalScopeContext(parent_lex)

    for value_info in graph.input:
        # TODO: If shadowing isn't allowed, this should maybe use
        # this_or_ancestor_graph_has
        if lex_ctx.this_graph_has(value_info.name):
            raise OnnxCheckError(  # pragma: no cover
                f"Graph must be in single static assignment (SSA) form, "
                f"however '{value_info.name}' has been used as "
                f"graph input names multiple times.",
                graph)
        lex_ctx.add(value_info.name)

    initializer_name_checker = set()
    # std::unordered_set<std::reference_wrapper<const std::string>, std::hash<std::string>, std::equal_to<std::string>>

    for init in graph.initializer:
        _enforce_has_field(init, "name")
        name = init.name
        if not name:
            raise OnnxCheckError(  # pragma: no cover
                f"Tensor initializers must have a non-empty name.",
                graph)

        if name in initializer_name_checker:
            raise OnnxCheckError(  # pragma: no cover
                f"'{name}' initializer name is not unique.",
                graph)
        initializer_name_checker.add(name)

        _check_tensor(init, ctx)

        if ctx.get_ir_version() <= 0x00000003:
            # Initializers are a subset of graph inputs for IR_VERSION <= 3
            if not lex_ctx.this_graph_has(name):
                raise OnnxCheckError(  # pragma: no cover
                    f"'{name}' in initializer but not in graph input.",
                    graph)
        else:
            # An initializer is allowed to have the same name as an input,
            # but is not required to (for IR_VERSION >= 4)
            lex_ctx.add(name)

    for sparse_init in graph.sparse_initializer:
        values = sparse_init.values()
        _enforce_has_field(values, name)
        name = values.name
        if name.empty():
            raise OnnxCheckError(  # pragma: no cover
                f"Sparse tensor initializers must have a non-empty name.",
                graph)
        if name in initializer_name_checker:
            raise OnnxCheckError(  # pragma: no cover
                f"'{name}' initializer name is not unique across "
                f"initializers and sparse_initializers.",
                graph)
        initializer_name_checker.add(name)
        _check_sparse_tensor(sparse_init, ctx)
        lex_ctx.add(name)

    errors = []
    for node in graph.node:
        # nodes must be in topologically sorted order
        for input in node.input:
            # explicit optional input
            if not input:
                continue
            if not lex_ctx.this_or_ancestor_graph_has(input):
                raise OnnxCheckError(  # pragma: no cover
                    f"Nodes in a graph must be topologically sorted, however "
                    f"input '{input}' of node name '{node.name}', type "
                    f"'{node.op_type}' is not output of any previous nodes.",
                    node)

        # This needs to happen before SSA check since we don't want to recurse and
        # find that outputs from control flow ops are colliding with names in the
        # inner block

        try:
            _check_node(node, ctx, lex_ctx)
        except OnnxCheckError as e:
            errors.append(e)

        # check for SSA form
        for output in node.output:
            # optional output
            if not output:
                continue

            if lex_ctx.this_or_ancestor_graph_has(output):
                raise OnnxCheckError(  # pragma: no cover
                    f"Graph must be in single static assignment "
                    f"(SSA) form, however '{output}' "
                    f"has been used as output names multiple times.",
                    graph)
            lex_ctx.add(output)


def _get_version_for_domain(domain, opset_imports):
    # Utilify function to get the imported version of domain from opset imports
    # Returns -1 if requested domain is not found in the opset_imports
    if domain not in opset_imports.end():
        return -1
    return opset_imports[domain]


def _check_opset_compatibility(node, ctx, func_opset_imports, model_opset_imports):
    func_opset_version = _get_version_for_domain(
        node.domain, func_opset_imports)
    model_opset_version = _get_version_for_domain(
        node.domain, model_opset_imports)

    if func_opset_version == -1:
        raise OnnxCheckError(  # pragma: no cover
            f"No Opset registered for domain '{node.domain}'.",
            node)

    if model_opset_version == -1:
        # model does not include opset import for a node present in function body.
        # This is ok as along as the opset import is present in function level opset imports.
        return

    if func_opset_version == model_opset_version:
        # both versions are same, no need to verify schema.
        return

    schema_for_model_import = ctx.get_schema_registry().GetSchema(
        node.op_type, model_opset_version, node.domain)
    schema_for_function_import = ctx.get_schema_registry().GetSchema(
        node.op_type, func_opset_version, node.domain)

    if not schema_for_model_import and not schema_for_function_import:
        # the op belongs to a custom domain so we cannot verify schema
        return

    # if schema is present for 1 but not other or the schema since
    # versions do not match then raise an error
    if (not schema_for_model_import or not schema_for_function_import or
            schema_for_function_import.since_version() != schema_for_model_import.since_version()):
        raise OnnxCheckError(  # pragma: no cover
            f"Opset import for domain '{node.domain}' in function op "
            f"'{node.op_type} is not compatible with the version "
            f"imported by model. FunctionOp imports version "
            f"{func_opset_version} whereas model imports version "
            f"{model_opset_version}.",
            node)


def _check_model_local_functions(model, ctx, parent_lex):
    # make a copy of model opset imports to maintain a main copy of opset imports across the model and
    # all model local functions to verify opset compatibility
    model_opset_imports = ctx.get_opset_imports()

    # merge the opset imports from every function in model_opset_imports
    # only add the opset import if an entry for it does not exist in model_opset_imports
    # if there is an entry then the compatibility will be checked later
    # on in check_opset_compatibility
    # called by check_function.
    for function_proto in model.functions:
        for opset_import in function_proto.opset_import():
            if _get_version_for_domain(opset_import.domain, model_opset_imports) == -1:
                model_opset_imports[opset_import.domain] = opset_import.version

    ctx_copy = CheckerContext(ctx)
    ctx_copy.set_opset_imports(model_opset_imports)

    for function_proto in model.functions:
        _check_function(function_proto, ctx_copy, parent_lex)


def _check_function(function, ctx, parent_lex):
    _enforce_non_empty_field(function, "name")

    if ctx.get_ir_version() >= 0x00000008:
        _enforce_has_field(function, "domain")

    model_opset_imports = ctx.get_opset_imports()
    ctx_copy = CheckerContext(ctx)

    func_opset_imports = {}
    for relied_opset in function.opset_import():
        func_opset_imports[relied_opset.domain] = int(relied_opset.version)

    ctx_copy.set_opset_imports(func_opset_imports)

    lex_ctx = LexicalScopeContext(parent_lex)

    for input in function.input:
        # TODO: If shadowing isn't allowed, this should maybe use
        # this_or_ancestor_graph_has
        if lex_ctx.this_graph_has(input):
            raise OnnxCheckError(  # pragma: no cover
                f"Graph must be in single static assignment (SSA) form, "
                f"however '{input}' has been used multiple times.",
                function)
        lex_ctx.add(input)

    outputs = set()
    for output in function.output:
        if output in outputs:
            raise OnnxCheckError(  # pragma: no cover
                f"Function '{function.name}' should not have "
                f"duplicate outputs specified.",
                function)
        outputs.add(output)

    attrs = set()
    for attr in function.attribute:
        if attr in attrs:
            raise OnnxCheckError(  # pragma: no cover
                f"Function '{function.name}' should not have "
                f"duplicate attributes specified.",
                function)

    for node in function.node():
        # nodes must be in topologically sorted order
        for input in node.input:
            # explicit optional input
            if input.empty():
                continue
            if not lex_ctx.this_graph_has(input):
                raise OnnxCheckError(  # pragma: no cover
                    f"Nodes in a function must be topologically sorted, "
                    f"however input '{input}' of node name '{node.name}' "
                    f"and type '{node.op_type}' is neither output "
                    f"of any previous nodes nor input of the function.",
                    function)

        # check whether the opset version imported for a domain by function and model are
        # compatible
        _check_opset_compatibility(
            node, ctx_copy, func_opset_imports, model_opset_imports)
        _check_node(node, ctx_copy, lex_ctx)

        # check for SSA form
        for output in node.output:
            # optional output
            if output.empty():
                continue

            if lex_ctx.this_or_ancestor_graph_has(output):
                raise OnnxCheckError(  # pragma: no cover
                    f"Function must be in single static assignment (SSA) "
                    f"form, however '{output}' has been used as output "
                    f"names multiple times.",
                    function)
            lex_ctx.add(output)


def _check_model(model, ctx):
    if not model.ir_version:
        raise OnnxCheckError(  # pragma: no cover
            f"The model does not have an ir_version set properly.",
            model)
    if model.ir_version > IR_VERSION:
        raise OnnxCheckError(  # pragma: no cover
            f"Your model ir_version is higher than the checker's.",
            model)
    if len(model.metadata_props) > 1:
        keys = set()
        for entry in model.metadata_props:
            if entry.key() in keys:
                raise OnnxCheckError(  # pragma: no cover
                    f"Your model has duplicate keys '{entry.key()}' "
                    f"in metadata_props.", model)
            keys.add(entry.key())

    ctx.set_ir_version(int(model.ir_version))
    opset_imports = {}
    for opset_import in model.opset_import:
        opset_imports[opset_import.domain] = int(opset_import.version)
    if model.ir_version >= 3:
        if not opset_imports:
            raise OnnxCheckError(  # pragma: no cover
                f"Model with IR version >= 3 must specify opset_import for "
                f"ONNX ({opset_imports}).",
                model)
    elif not opset_imports:
        opset_imports[ONNX_DOMAIN] = 1
    else:
        raise OnnxCheckError(  # pragma: no cover
            f"Model with IR version < 3 cannot have opset_import specified.",
            model)

    ctx.set_opset_imports(opset_imports)
    lex_ctx = LexicalScopeContext()
    _check_graph(model.graph, ctx, lex_ctx)

    if ctx.get_ir_version() >= 0x00000008:
        _check_model_local_functions(model, ctx, lex_ctx)


def check_model(model):
    """
    Checks a model is consistent with ONNX language.
    The function fails if the model is not consistent.

    :param model: :epkg:`ModelProto`
    """
    ctx = CheckerContext()
    if isinstance(model, bytes):
        m = ModelProto()
        m.ParseFromString(model)
        _check_model(m, ctx)
    else:
        _check_model(model, ctx)


experimental_ops = {
    "ATen",
    "Affine",
    "ConstantFill",
    "Crop",
    "DynamicSlice",
    "GRUUnit",
    "GivenTensorFill",
    "ImageScaler",
    "ParametricSoftplus",
    "Scale",
    "ScaledTanh"}


def check_is_experimental_op(node_op_type):
    "Tells if an operator is experimentation."
    return bool(experimental_ops & {node_op_type})



"""
@file
@brief Rewrites some of the converters implemented in
:epkg:`sklearn-onnx`.
"""
from skl2onnx.common._registration import (
    _converter_pool, _shape_calculator_pool)
try:
    from skl2onnx.common._registration import RegisteredConverter
except ImportError:  # pragma: no cover
    # sklearn-onnx <= 1.6.0
    RegisteredConverter = lambda fct, opts: fct
from .sklconv.tree_converters import (
    new_convert_sklearn_decision_tree_classifier,
    new_convert_sklearn_decision_tree_regressor,
    new_convert_sklearn_gradient_boosting_classifier,
    new_convert_sklearn_gradient_boosting_regressor,
    new_convert_sklearn_random_forest_classifier,
    new_convert_sklearn_random_forest_regressor)
from .sklconv.svm_converters import (
    new_convert_sklearn_svm_classifier,
    new_convert_sklearn_svm_regressor)
from .sklconv.function_transformer_converters import (
    new_calculate_sklearn_function_transformer_output_shapes,
    new_convert_sklearn_function_transformer)


_overwritten_operators = {
    #
    'SklearnOneClassSVM': RegisteredConverter(
        new_convert_sklearn_svm_regressor,
        _converter_pool['SklearnOneClassSVM'].get_allowed_options()),
    'SklearnSVR': RegisteredConverter(
        new_convert_sklearn_svm_regressor,
        _converter_pool['SklearnSVR'].get_allowed_options()),
    'SklearnSVC': RegisteredConverter(
        new_convert_sklearn_svm_classifier,
        _converter_pool['SklearnSVC'].get_allowed_options()),
    #
    'SklearnDecisionTreeRegressor': RegisteredConverter(
        new_convert_sklearn_decision_tree_regressor,
        _converter_pool['SklearnDecisionTreeRegressor'].get_allowed_options()),
    'SklearnDecisionTreeClassifier': RegisteredConverter(
        new_convert_sklearn_decision_tree_classifier,
        _converter_pool['SklearnDecisionTreeClassifier'].get_allowed_options()),
    #
    'SklearnExtraTreeRegressor': RegisteredConverter(
        new_convert_sklearn_decision_tree_regressor,
        _converter_pool['SklearnExtraTreeRegressor'].get_allowed_options()),
    'SklearnExtraTreeClassifier': RegisteredConverter(
        new_convert_sklearn_decision_tree_classifier,
        _converter_pool['SklearnExtraTreeClassifier'].get_allowed_options()),
    #
    'SklearnExtraTreesRegressor': RegisteredConverter(
        new_convert_sklearn_random_forest_regressor,
        _converter_pool['SklearnExtraTreesRegressor'].get_allowed_options()),
    'SklearnExtraTreesClassifier': RegisteredConverter(
        new_convert_sklearn_random_forest_classifier,
        _converter_pool['SklearnExtraTreesClassifier'].get_allowed_options()),
    #
    'SklearnFunctionTransformer': RegisteredConverter(
        new_convert_sklearn_function_transformer,
        _converter_pool['SklearnFunctionTransformer'].get_allowed_options()),
    #
    'SklearnGradientBoostingRegressor': RegisteredConverter(
        new_convert_sklearn_gradient_boosting_regressor,
        _converter_pool['SklearnGradientBoostingRegressor'].get_allowed_options()),
    'SklearnGradientBoostingClassifier': RegisteredConverter(
        new_convert_sklearn_gradient_boosting_classifier,
        _converter_pool['SklearnGradientBoostingClassifier'].get_allowed_options()),
    #
    'SklearnHistGradientBoostingRegressor': RegisteredConverter(
        new_convert_sklearn_random_forest_regressor,
        _converter_pool['SklearnHistGradientBoostingRegressor'].get_allowed_options()),
    'SklearnHistGradientBoostingClassifier': RegisteredConverter(
        new_convert_sklearn_random_forest_classifier,
        _converter_pool['SklearnHistGradientBoostingClassifier'].get_allowed_options()),
    #
    'SklearnRandomForestRegressor': RegisteredConverter(
        new_convert_sklearn_random_forest_regressor,
        _converter_pool['SklearnRandomForestRegressor'].get_allowed_options()),
    'SklearnRandomForestClassifier': RegisteredConverter(
        new_convert_sklearn_random_forest_classifier,
        _converter_pool['SklearnRandomForestClassifier'].get_allowed_options()),
}

_overwritten_shape_calculator = {
    "SklearnFunctionTransformer":
        new_calculate_sklearn_function_transformer_output_shapes,
}


def register_rewritten_operators(new_converters=None,
                                 new_shape_calculators=None):
    """
    Registers modified operators and returns the old values.

    :param new_converters: converters to rewrite or None
        to rewrite default ones
    :param new_shape_calculators: shape calculators to rewrite or
        None to rewrite default ones
    @return old converters, old shape calculators
    """
    old_conv = None
    old_shape = None

    if new_converters is None:
        for rew in _overwritten_operators:
            if rew not in _converter_pool:
                raise KeyError(  # pragma: no cover
                    f"skl2onnx was not imported and '{rew}' was not registered.")
        old_conv = {k: _converter_pool[k] for k in _overwritten_operators}
        _converter_pool.update(_overwritten_operators)
    else:
        for rew in new_converters:
            if rew not in _converter_pool:
                raise KeyError(  # pragma: no cover
                    f"skl2onnx was not imported and '{rew}' was not registered.")
        old_conv = {k: _converter_pool[k] for k in new_converters}
        _converter_pool.update(new_converters)

    if new_shape_calculators is None:
        for rew in _overwritten_shape_calculator:
            if rew not in _shape_calculator_pool:
                raise KeyError(  # pragma: no cover
                    f"skl2onnx was not imported and '{rew}' was not registered.")
        old_shape = {k: _shape_calculator_pool[k]
                     for k in _overwritten_shape_calculator}
        _shape_calculator_pool.update(_overwritten_shape_calculator)
    else:
        for rew in new_shape_calculators:
            if rew not in _shape_calculator_pool:
                raise KeyError(  # pragma: no cover
                    f"skl2onnx was not imported and '{rew}' was not registered.")
        old_shape = {k: _shape_calculator_pool[k]
                     for k in new_shape_calculators}
        _shape_calculator_pool.update(new_shape_calculators)

    return old_conv, old_shape



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class CommonLSTM(OpRun):

    def __init__(self, onnx_node, expected_attributes=None, desc=None,
                 **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=expected_attributes,
                       **options)
        self.nb_outputs = len(onnx_node.output)
        self.number_of_gates = 3

    def f(self, x):
        return 1 / (1 + numpy.exp(-x))

    def g(self, x):
        return numpy.tanh(x)

    def h(self, x):
        return numpy.tanh(x)

    def _step(self, X, R, B, W, H_0, C_0, P):
        seq_length = X.shape[0]
        hidden_size = H_0.shape[-1]
        batch_size = X.shape[1]

        Y = numpy.empty(
            [seq_length, self.num_directions, batch_size, hidden_size])
        h_list = []

        [p_i, p_o, p_f] = numpy.split(P, 3)  # pylint: disable=W0632
        H_t = H_0
        C_t = C_0
        for x in numpy.split(X, X.shape[0], axis=0):
            gates = numpy.dot(x, numpy.transpose(W)) + numpy.dot(H_t, numpy.transpose(R)) + numpy.add(
                *numpy.split(B, 2))
            i, o, f, c = numpy.split(gates, 4, -1)  # pylint: disable=W0632
            i = self.f(i + p_i * C_t)
            f = self.f(f + p_f * C_t)
            c = self.g(c)
            C = f * C_t + i * c
            o = self.f(o + p_o * C)
            H = o * self.h(C)
            h_list.append(H)
            H_t = H
            C_t = C

        concatenated = numpy.concatenate(h_list)
        if self.num_directions == 1:
            Y[:, 0, :, :] = concatenated

        if self.layout == 0:
            Y_h = Y[-1]
        else:
            Y = numpy.transpose(Y, [2, 0, 1, 3])
            Y_h = Y[:, :, -1, :]

        return Y, Y_h

    def _run(self, X, W, R, B=None, sequence_lens=None,  # pylint: disable=W0221
             initial_h=None, initial_c=None, P=None,
             attributes=None, verbose=0, fLOG=None):
        number_of_gates = 4
        number_of_peepholes = 3

        self.num_directions = W.shape[0]

        if self.num_directions == 1:
            R = numpy.squeeze(R, axis=0)
            W = numpy.squeeze(W, axis=0)
            if B is not None:
                B = numpy.squeeze(B, axis=0)
            if sequence_lens is not None:
                sequence_lens = numpy.squeeze(sequence_lens, axis=0)
            if initial_h is not None:
                initial_h = numpy.squeeze(initial_h, axis=0)
            if initial_c is not None:
                initial_c = numpy.squeeze(initial_c, axis=0)
            if P is not None:
                P = numpy.squeeze(P, axis=0)

            hidden_size = R.shape[-1]
            batch_size = X.shape[1]

            if self.layout != 0:
                X = numpy.swapaxes(X, 0, 1)
            if B is None:
                B = numpy.zeros(2 * number_of_gates *
                                hidden_size, dtype=numpy.float32)
            if P is None:
                P = numpy.zeros(number_of_peepholes *
                                hidden_size, dtype=numpy.float32)
            if initial_h is None:
                initial_h = numpy.zeros(
                    (batch_size, hidden_size), dtype=numpy.float32)
            if initial_c is None:
                initial_c = numpy.zeros(
                    (batch_size, hidden_size), dtype=numpy.float32)
        else:
            raise NotImplementedError(  # pragma: no cover
                "Unsupported value %r for num_directions and operator %r." % (
                    self.num_directions, self.__class__.__name__))

        Y, Y_h = self._step(X, R, B, W, initial_h, initial_c, P)

        return (Y, ) if self.nb_outputs == 1 else (Y, Y_h)


class LSTM(CommonLSTM):

    atts = {
        'activation_alpha': [0.],
        'activation_beta': [0.],
        'activations': [b'Tanh', b'Tanh'],
        'clip': [],
        'direction': b'forward',
        'hidden_size': None,
        'layout': 0,
        'input_forget': 0,
    }

    def __init__(self, onnx_node, desc=None, **options):
        CommonLSTM.__init__(self, onnx_node, desc=desc,
                            expected_attributes=LSTM.atts,
                            **options)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class ConstantOfShape(OpRun):

    atts = {'value': numpy.array([0], dtype=numpy.float32)}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=ConstantOfShape.atts,
                       **options)
        self.cst = (self.value[0]
                    if isinstance(self.value, numpy.ndarray)
                    else self.value)
        if isinstance(self.cst, int):
            self.cst = numpy.int64(self.cst)
        elif isinstance(self.cst, float):
            self.cst = numpy.float64(self.cst)
        if not isinstance(self.cst, (numpy.float32, numpy.float64,
                                     numpy.int64, numpy.int32, numpy.bool_,
                                     numpy.float16)):
            raise TypeError(  # pragma: no cover
                f"cst must be a real not {type(self.cst)}")

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        try:
            res = numpy.full(tuple(data), self.cst)
        except TypeError as e:  # pragma: no cover
            raise RuntimeError(
                "Unable to create a constant of shape %r with value %r "
                "(raw value=%r)." % (data, self.cst, self.value)) from e
        return (res, )

    def to_python(self, inputs):
        lines = ['cst = value[0] if isinstance(value, numpy.ndarray) else value',
                 f'return numpy.full(tuple({inputs[0]}), cst)']
        return ("import numpy", "\n".join(lines))



"""
@file
@brief Helpers to speed up the conversion of Lightgbm models or transform it.
"""
from collections import deque
import ctypes
import json
import re


def restore_lgbm_info(tree):
    """
    Restores speed up information to help
    modifying the structure of the tree.
    """

    def walk_through(t):
        if 'tree_info' in t:
            yield None
        elif 'tree_structure' in t:
            for w in walk_through(t['tree_structure']):
                yield w
        else:
            yield t
            if 'left_child' in t:
                for w in walk_through(t['left_child']):
                    yield w
            if 'right_child' in t:
                for w in walk_through(t['right_child']):
                    yield w

    nodes = []
    if 'tree_info' in tree:
        for node in walk_through(tree):
            if node is None:
                nodes.append([])
            elif 'right_child' in node or 'left_child' in node:
                nodes[-1].append(node)
    else:
        for node in walk_through(tree):
            if 'right_child' in node or 'left_child' in node:
                nodes.append(node)
    return nodes


def dump_booster_model(self, num_iteration=None, start_iteration=0,
                       importance_type='split', verbose=0):
    """
    Dumps Booster to JSON format.

    Parameters
    ----------
    self: booster
    num_iteration : int or None, optional (default=None)
        Index of the iteration that should be dumped.
        If None, if the best iteration exists, it is dumped; otherwise,
        all iterations are dumped.
        If <= 0, all iterations are dumped.
    start_iteration : int, optional (default=0)
        Start index of the iteration that should be dumped.
    importance_type : string, optional (default="split")
        What type of feature importance should be dumped.
        If "split", result contains numbers of times the feature is used in a model.
        If "gain", result contains total gains of splits which use the feature.
    verbose: dispays progress (usefull for big trees)

    Returns
    -------
    json_repr : dict
        JSON format of Booster.

    .. note::
        This function is inspired from
        the :epkg:`lightgbm` (`dump_model
        <https://lightgbm.readthedocs.io/en/latest/pythonapi/
        lightgbm.Booster.html#lightgbm.Booster.dump_model>`_.
        It creates intermediate structure to speed up the conversion
        into ONNX of such model. The function overwrites the
        `json.load` to fastly extract nodes.
    """
    if getattr(self, 'is_mock', False):
        return self.dump_model(), None
    from lightgbm.basic import (
        _LIB, FEATURE_IMPORTANCE_TYPE_MAPPER, _safe_call,
        json_default_with_numpy)
    if num_iteration is None:
        num_iteration = self.best_iteration
    importance_type_int = FEATURE_IMPORTANCE_TYPE_MAPPER[importance_type]
    buffer_len = 1 << 20
    tmp_out_len = ctypes.c_int64(0)
    string_buffer = ctypes.create_string_buffer(buffer_len)
    ptr_string_buffer = ctypes.c_char_p(*[ctypes.addressof(string_buffer)])
    if verbose >= 2:
        print(  # pragma: no cover
            "[dump_booster_model] call CAPI: LGBM_BoosterDumpModel")
    _safe_call(_LIB.LGBM_BoosterDumpModel(
        self.handle,
        ctypes.c_int(start_iteration),
        ctypes.c_int(num_iteration),
        ctypes.c_int(importance_type_int),
        ctypes.c_int64(buffer_len),
        ctypes.byref(tmp_out_len),
        ptr_string_buffer))
    actual_len = tmp_out_len.value
    # if buffer length is not long enough, reallocate a buffer
    if actual_len > buffer_len:
        string_buffer = ctypes.create_string_buffer(actual_len)
        ptr_string_buffer = ctypes.c_char_p(
            *[ctypes.addressof(string_buffer)])
        _safe_call(_LIB.LGBM_BoosterDumpModel(
            self.handle,
            ctypes.c_int(start_iteration),
            ctypes.c_int(num_iteration),
            ctypes.c_int(importance_type_int),
            ctypes.c_int64(actual_len),
            ctypes.byref(tmp_out_len),
            ptr_string_buffer))

    WHITESPACE = re.compile(
        r'[ \t\n\r]*', re.VERBOSE | re.MULTILINE | re.DOTALL)

    class Hook(json.JSONDecoder):
        """
        Keep track of the progress, stores a copy of all objects with
        a decision into a different container in order to walk through
        all nodes in a much faster way than going through the architecture.
        """

        def __init__(self, *args, info=None, n_trees=None, verbose=0,
                     **kwargs):
            json.JSONDecoder.__init__(
                self, object_hook=self.hook, *args, **kwargs)
            self.nodes = []
            self.buffer = []
            self.info = info
            self.n_trees = n_trees
            self.verbose = verbose
            self.stored = 0
            if verbose >= 2 and n_trees is not None:
                from tqdm import tqdm  # pragma: no cover
                self.loop = tqdm(total=n_trees)  # pragma: no cover
                self.loop.set_description("dump_booster")  # pragma: no cover
            else:
                self.loop = None

        def decode(self, s, _w=WHITESPACE.match):
            return json.JSONDecoder.decode(self, s, _w=_w)

        def raw_decode(self, s, idx=0):
            return json.JSONDecoder.raw_decode(self, s, idx=idx)

        def hook(self, obj):
            """
            Hook called everytime a JSON object is created.
            Keep track of the progress, stores a copy of all objects with
            a decision into a different container.
            """
            # Every obj goes through this function from the leaves to the root.
            if 'tree_info' in obj:
                self.info['decision_nodes'] = self.nodes
                if self.n_trees is not None and len(self.nodes) != self.n_trees:
                    raise RuntimeError(  # pragma: no cover
                        "Unexpected number of trees %d (expecting %d)." % (
                            len(self.nodes), self.n_trees))
                self.nodes = []
                if self.loop is not None:
                    self.loop.close()
            if 'tree_structure' in obj:
                self.nodes.append(self.buffer)
                if self.loop is not None:
                    self.loop.update(len(self.nodes))
                    if len(self.nodes) % 10 == 0:
                        self.loop.set_description(
                            "dump_booster: %d/%d trees, %d nodes" % (
                                len(self.nodes), self.n_trees, self.stored))
                self.buffer = []
            if "decision_type" in obj:
                self.buffer.append(obj)
                self.stored += 1
            return obj

    if verbose >= 2:
        print("[dump_booster_model] to_json")  # pragma: no cover
    info = {}
    ret = json.loads(string_buffer.value.decode('utf-8'), cls=Hook,
                     info=info, n_trees=self.num_trees(), verbose=verbose)
    ret['pandas_categorical'] = json.loads(
        json.dumps(self.pandas_categorical,
                   default=json_default_with_numpy))
    if verbose >= 2:
        print("[dump_booster_model] end.")  # pragma: no cover
    return ret, info


def dump_lgbm_booster(booster, verbose=0):
    """
    Dumps a Lightgbm booster into JSON.

    :param booster: Lightgbm booster
    :param verbose: verbosity
    :return: json, dictionary with more information
    """
    js, info = dump_booster_model(booster, verbose=verbose)
    return js, info


def modify_tree_for_rule_in_set(gbm, use_float=False, verbose=0, count=0,  # pylint: disable=R1710
                                info=None):
    """
    LightGBM produces sometimes a tree with a node set
    to use rule ``==`` to a set of values (= in set),
    the values are separated by ``||``.
    This function unfold theses nodes.

    :param gbm: a tree coming from lightgbm dump
    :param use_float: use float otherwise int first
        then float if it does not work
    :param verbose: verbosity, use :epkg:`tqdm` to show progress
    :param count: number of nodes already changed (origin) before this call
    :param info: addition information to speed up this search
    :return: number of changed nodes (include *count*)

    A child looks like the following:

    .. runpython::
        :showcode:
        :warningout: DeprecationWarning

        import pprint
        from mlprodict.onnx_conv.operator_converters.conv_lightgbm import modify_tree_for_rule_in_set

        tree = {'decision_type': '==',
                'default_left': True,
                'internal_count': 6805,
                'internal_value': 0.117558,
                'left_child': {'leaf_count': 4293,
                               'leaf_index': 18,
                               'leaf_value': 0.003519117642745049},
                'missing_type': 'None',
                'right_child': {'leaf_count': 2512,
                                'leaf_index': 25,
                                'leaf_value': 0.012305307958365394},
                'split_feature': 24,
                'split_gain': 12.233599662780762,
                'split_index': 24,
                'threshold': '10||12||13'}

        modify_tree_for_rule_in_set(tree)

        pprint.pprint(tree)
    """
    if 'tree_info' in gbm:
        if info is not None:
            dec_nodes = info['decision_nodes']
        else:
            dec_nodes = None
        if verbose >= 2:  # pragma: no cover
            from tqdm import tqdm
            loop = tqdm(gbm['tree_info'])
            for i, tree in enumerate(loop):
                loop.set_description("rules tree %d c=%d" % (i, count))
                count = modify_tree_for_rule_in_set(
                    tree, use_float=use_float, count=count,
                    info=None if dec_nodes is None else dec_nodes[i])
        else:
            for i, tree in enumerate(gbm['tree_info']):
                count = modify_tree_for_rule_in_set(
                    tree, use_float=use_float, count=count,
                    info=None if dec_nodes is None else dec_nodes[i])
        return count

    if 'tree_structure' in gbm:
        return modify_tree_for_rule_in_set(
            gbm['tree_structure'], use_float=use_float, count=count,
            info=info)

    if 'decision_type' not in gbm:
        return count

    def str2number(val):
        if use_float:
            return float(val)
        else:
            try:
                return int(val)
            except ValueError:  # pragma: no cover
                return float(val)

    if info is None:

        def recursive_call(this, c):
            if 'left_child' in this:
                c = process_node(this['left_child'], count=c)
            if 'right_child' in this:
                c = process_node(this['right_child'], count=c)
            return c

        def process_node(node, count):
            if 'decision_type' not in node:
                return count
            if node['decision_type'] != '==':
                return recursive_call(node, count)
            th = node['threshold']
            if not isinstance(th, str):
                return recursive_call(node, count)
            pos = th.find('||')
            if pos == -1:
                return recursive_call(node, count)
            th1 = str2number(th[:pos])

            def doit():
                rest = th[pos + 2:]
                if '||' not in rest:
                    rest = str2number(rest)

                node['threshold'] = th1
                new_node = node.copy()
                node['right_child'] = new_node
                new_node['threshold'] = rest

            doit()
            return recursive_call(node, count + 1)

        return process_node(gbm, count)

    # when info is used

    def split_node(node, th, pos):
        th1 = str2number(th[:pos])

        rest = th[pos + 2:]
        if '||' not in rest:
            rest = str2number(rest)
            app = False
        else:
            app = True

        node['threshold'] = th1
        new_node = node.copy()
        node['right_child'] = new_node
        new_node['threshold'] = rest
        return new_node, app

    stack = deque(info)
    while len(stack) > 0:
        node = stack.pop()

        if 'decision_type' not in node:
            continue  # leave

        if node['decision_type'] != '==':
            continue

        th = node['threshold']
        if not isinstance(th, str):
            continue

        pos = th.find('||')
        if pos == -1:
            continue

        new_node, app = split_node(node, th, pos)
        count += 1
        if app:
            stack.append(new_node)

    return count



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from numpy.fft import fft2
from ._op import OpRun
from ._new_ops import OperatorSchema


class FFT2D(OpRun):

    atts = {'axes': [-2, -1]}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=FFT2D.atts,
                       **options)
        if self.axes is not None:
            self.axes = tuple(self.axes)
            if len(self.axes) != 2:
                raise ValueError(  # pragma: no cover
                    f"axes must a set of 1 integers not {self.axes!r}.")

    def _find_custom_operator_schema(self, op_name):
        if op_name == "FFT2D":
            return FFT2DSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def _run(self, a, fft_length=None, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if fft_length is None:
            y = fft2(a, axes=self.axes)
        else:
            y = fft2(a, tuple(fft_length), axes=self.axes)
        if a.dtype in (numpy.float32, numpy.complex64):
            return (y.astype(numpy.complex64), )
        if a.dtype in (numpy.float64, numpy.complex128):
            return (y.astype(numpy.complex128), )
        raise TypeError(  # pragma: no cover
            f"Unexpected input type: {a.dtype!r}.")

    def to_python(self, inputs):
        if self.axes is not None:
            axes = tuple(self.axes)
        else:
            axes = None
        if len(inputs) == 1:
            return ('from numpy.fft import fft2',
                    f"return fft2({inputs[0]}, axes={axes})")
        return ('from numpy.fft import fft2',
                f"return fft2({inputs[0]}, tuple({inputs[1]}), axes={axes})")


class FFT2DSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl FFT.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'FFT2D')
        self.attributes = FFT2D.atts



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class Gemm(OpRun):

    atts = {'alpha': 1., 'beta': 1., 'transA': 0, 'transB': 0}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Gemm.atts,
                       **options)
        if self.transA:
            _meth = (Gemm._gemm11 if self.transB
                     else Gemm._gemm10)
        else:
            _meth = (Gemm._gemm01 if self.transB
                     else Gemm._gemm00)
        self._meth = lambda a, b, c: _meth(a, b, c, self.alpha, self.beta)

    @staticmethod
    def _gemm00(a, b, c, alpha, beta):
        o = numpy.dot(a, b) * alpha
        if c is not None and beta != 0:
            o += c * beta
        return o

    @staticmethod
    def _gemm01(a, b, c, alpha, beta):
        o = numpy.dot(a, b.T) * alpha
        if c is not None and beta != 0:
            o += c * beta
        return o

    @staticmethod
    def _gemm10(a, b, c, alpha, beta):
        o = numpy.dot(a.T, b) * alpha
        if c is not None and beta != 0:
            o += c * beta
        return o

    @staticmethod
    def _gemm11(a, b, c, alpha, beta):
        o = numpy.dot(a.T, b.T) * alpha
        if c is not None and beta != 0:
            o += c * beta
        return o

    def _run(self, a, b, c=None, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (self._meth(a, b, c), )



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunBinaryNumpy


class Min(OpRunBinaryNumpy):

    def __init__(self, onnx_node, desc=None, **options):
        OpRunBinaryNumpy.__init__(self, numpy.minimum, onnx_node,
                                  desc=desc, **options)

    def run(self, *data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(data) == 2:
            return OpRunBinaryNumpy.run(self, *data, verbose=verbose, fLOG=fLOG)
        if len(data) == 1:
            if self.inplaces.get(0, False):
                return (data[0], )
            return (data[0].copy(), )
        if len(data) > 2:
            a = data[0]
            for i in range(1, len(data)):
                a = numpy.minimum(a, data[i])
            return (a, )
        raise RuntimeError(  # pragma: no cover
            "Unexpected turn of events.")



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class CommonGRU(OpRun):

    def __init__(self, onnx_node, expected_attributes=None, desc=None,
                 **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=expected_attributes,
                       **options)
        self.nb_outputs = len(onnx_node.output)
        self.number_of_gates = 3

    def f(self, x):
        return 1 / (1 + numpy.exp(-x))

    def g(self, x):
        return numpy.tanh(x)

    def _step(self, X, R, B, W, H_0):
        seq_length = X.shape[0]
        hidden_size = H_0.shape[-1]
        batch_size = X.shape[1]

        Y = numpy.empty(
            [seq_length, self.num_directions, batch_size, hidden_size])
        h_list = []

        [w_z, w_r, w_h] = numpy.split(W, 3)  # pylint: disable=W0632
        [r_z, r_r, r_h] = numpy.split(R, 3)  # pylint: disable=W0632
        [w_bz, w_br, w_bh, r_bz, r_br, r_bh] = numpy.split(  # pylint: disable=W0632
            B, 6)  # pylint: disable=W0632
        gates_w = numpy.transpose(numpy.concatenate((w_z, w_r)))
        gates_r = numpy.transpose(numpy.concatenate((r_z, r_r)))
        gates_b = numpy.add(numpy.concatenate((w_bz, w_br)),
                            numpy.concatenate((r_bz, r_br)))

        H_t = H_0
        for x in numpy.split(X, X.shape[0], axis=0):
            gates = numpy.dot(x, gates_w) + numpy.dot(H_t, gates_r) + gates_b
            z, r = numpy.split(gates, 2, -1)  # pylint: disable=W0632
            z = self.f(z)
            r = self.f(r)
            h_default = self.g(numpy.dot(x, numpy.transpose(
                w_h)) + numpy.dot(r * H_t, numpy.transpose(r_h)) + w_bh + r_bh)
            h_linear = self.g(numpy.dot(x, numpy.transpose(
                w_h)) + r * (numpy.dot(H_t, numpy.transpose(r_h)) + r_bh) + w_bh)
            h = h_linear if self.linear_before_reset else h_default
            H = (1 - z) * h + z * H_t
            h_list.append(H)
            H_t = H

        concatenated = numpy.concatenate(h_list)
        if self.num_directions == 1:
            Y[:, 0, :, :] = concatenated

        if self.layout == 0:
            Y_h = Y[-1]
        else:
            Y = numpy.transpose(Y, [2, 0, 1, 3])
            Y_h = Y[:, :, -1, :]

        return Y, Y_h

    def _run(self, X, W, R, B=None, attributes=None, sequence_lens=None,  # pylint: disable=W0221
             initial_h=None, verbose=0, fLOG=None):
        self.num_directions = W.shape[0]

        if self.num_directions == 1:
            R = numpy.squeeze(R, axis=0)
            W = numpy.squeeze(W, axis=0)
            if B is not None:
                B = numpy.squeeze(B, axis=0)
            if sequence_lens is not None:
                sequence_lens = numpy.squeeze(sequence_lens, axis=0)
            if initial_h is not None:
                initial_h = numpy.squeeze(initial_h, axis=0)

            hidden_size = R.shape[-1]
            batch_size = X.shape[1]

            b = (B if B is not None else
                 numpy.zeros(2 * self.number_of_gates * hidden_size, dtype=X.dtype))
            h_0 = (initial_h if initial_h is not None else
                   numpy.zeros((batch_size, hidden_size), dtype=X.dtype))

            B = b
            H_0 = h_0
        else:
            raise NotImplementedError(  # pragma: no cover
                "Unsupported value %r for num_directions and operator %r." % (
                    self.num_directions, self.__class__.__name__))

        Y, Y_h = self._step(X, R, B, W, H_0)

        return (Y, ) if self.nb_outputs == 1 else (Y, Y_h)


class GRU(CommonGRU):

    atts = {
        'activation_alpha': [0.],
        'activation_beta': [0.],
        'activations': [b'Tanh', b'Tanh'],
        'clip': [],
        'direction': b'forward',
        'hidden_size': None,
        'layout': 0,
        'linear_before_reset': 0,
    }

    def __init__(self, onnx_node, desc=None, **options):
        CommonGRU.__init__(self, onnx_node, desc=desc,
                           expected_attributes=GRU.atts,
                           **options)



# coding: utf-8
"""
@file
@brief Wraps runtime into a :epkg:`scikit-learn` transformer.
"""
from io import BytesIO
import numpy
import pandas
import onnx
from sklearn.base import BaseEstimator, TransformerMixin
from skl2onnx.algebra.onnx_operator_mixin import OnnxOperatorMixin
from mlprodict.onnx_tools.onnx_manipulations import (
    select_model_inputs_outputs, enumerate_model_node_outputs)
from ..onnx_tools.onnx2py_helper import _var_as_dict, onnx_model_opsets
from ..onnx_tools.exports.skl2onnx_helper import add_onnx_graph
from ..onnxrt import OnnxInference


class OnnxTransformer(BaseEstimator, TransformerMixin, OnnxOperatorMixin):
    """
    Calls :epkg:`onnxruntime` or the runtime implemented
    in this package to transform input based on a ONNX graph.
    It follows :epkg:`scikit-learn` API
    so that it can be included in a :epkg:`scikit-learn` pipeline.
    See notebook :ref:`transferlearningrst` for an example.

    :param onnx_bytes: bytes
    :param output_name: string
        requested output name or None to request all and
        have method *transform* to store all of them in a dataframe
    :param enforce_float32: boolean
        :epkg:`onnxruntime` only supports *float32*,
        :epkg:`scikit-learn` usually uses double floats, this parameter
        ensures that every array of double floats is converted into
        single floats
    :param runtime: string, defined the runtime to use
        as described in @see cl OnnxInference.
    :param change_batch_size: some models are converted for
        a specific batch size, this parameter changes it,
        None to avoid changing it, 0 to fix an undefined
        first dimension
    :param reshape: reshape the output to get
        a matrix and not a multidimensional array
    """

    def __init__(self, onnx_bytes, output_name=None, enforce_float32=True,
                 runtime='python', change_batch_size=None, reshape=False):
        BaseEstimator.__init__(self)
        TransformerMixin.__init__(self)
        self.onnx_bytes = (onnx_bytes
                           if not hasattr(onnx_bytes, 'SerializeToString')
                           else onnx_bytes.SerializeToString())
        self.output_name = output_name
        self.enforce_float32 = enforce_float32
        self.runtime = runtime
        self.change_batch_size = change_batch_size
        self.reshape = reshape

    def __repr__(self):  # pylint: disable=W0222
        """
        usual
        """
        ob = self.onnx_bytes
        if len(ob) > 20:
            ob = ob[:10] + b"..." + ob[-10:]
        return ("{0}(onnx_bytes={1}, output_name={2}, enforce_float32={3}, "
                "runtime='{4}')".format(
                    self.__class__.__name__, ob, self.output_name,
                    self.enforce_float32, self.runtime))

    def fit(self, X=None, y=None, **fit_params):
        """
        Loads the :epkg:`ONNX` model.

        :param X: unused
        :param y: unused
        :param fit_params: additional parameter (unused)
        :return: self
        """
        from ..onnx_tools.optim.onnx_helper import change_input_first_dimension
        onx = onnx.load(BytesIO(self.onnx_bytes))
        self.op_version = onnx_model_opsets(onx)

        output_names = set(
            o.name for o in onx.graph.output)  # pylint: disable=E1101
        updated = False
        if (self.output_name is not None and
                self.output_name not in output_names):
            # The model refers to intermediate outputs.
            onx = select_model_inputs_outputs(
                onx, outputs=[self.output_name])
            updated = True

        if self.change_batch_size is not None:
            onx = change_input_first_dimension(
                onx, self.change_batch_size)
            updated = True

        onnx_bytes = (
            onx.SerializeToString() if updated else self.onnx_bytes)
        self.onnxrt_ = OnnxInference(
            onnx_bytes, runtime=self.runtime,
            runtime_options=dict(log_severity_level=3))
        self.inputs_ = self.onnxrt_.input_names
        self.inputs_shape_types_ = self.onnxrt_.input_names_shapes_types
        return self

    def _check_arrays(self, inputs):
        """
        Ensures that double floats are converted into single floats
        if *enforce_float32* is True or raises an exception.
        """
        has = hasattr(self, "onnxrt_")
        sht = self.inputs_shape_types_ if has else None
        if sht is not None and len(sht) < len(inputs):
            raise RuntimeError(  # pragma: no cover
                f"Unexpected number of inputs {len(inputs)} > {len(sht)} (expected).")
        for i, k in enumerate(inputs):
            v = inputs[k]
            if isinstance(v, numpy.ndarray):
                if v.dtype == numpy.float64 and self.enforce_float32:
                    inputs[k] = v.astype(numpy.float32)
                    continue
                if not has:
                    continue
                exp = sht[i]
                if exp[1] != ('?', ) and exp[1][1:] != v.shape[1:]:
                    raise RuntimeError(  # pragma: no cover
                        "Unexpected shape for input '{}': {} != {} "
                        "(expected).".format(
                            k, v.shape, exp[1]))
                if ((v.dtype == numpy.float32 and exp[2] != 'tensor(float)') or
                        (v.dtype == numpy.float64 and exp[2] != 'tensor(double)')):
                    raise TypeError(  # pragma: no cover
                        "Unexpected dtype for input '{}': {} != {} "
                        "(expected).".format(
                            k, v.dtype, exp[2]))

    def transform(self, X, y=None, **inputs):
        """
        Runs the predictions. If *X* is a dataframe,
        the function assumes every columns is a separate input,
        otherwise, *X* is considered as a first input and *inputs*
        can be used to specify extra inputs.

        :param X: iterable, data to process
            (or first input if several expected)
        :param y: unused
        :param inputs: :epkg:`ONNX` graph support multiple inputs,
            each column of a dataframe is converted into as many inputs if
            *X* is a dataframe, otherwise, *X* is considered as the first input
            and *inputs* can be used to specify the other ones
        :return: :epkg:`DataFrame`
        """
        if not hasattr(self, "onnxrt_"):
            raise AttributeError(  # pragma: no cover
                "Transform OnnxTransformer must be fit first.")
        rt_inputs = {}
        if isinstance(X, numpy.ndarray):
            rt_inputs[self.inputs_[0]] = X
        elif isinstance(X, pandas.DataFrame):
            for c in X.columns:
                rt_inputs[c] = X[c]
        elif isinstance(X, dict) and len(inputs) == 0:
            for k, v in X.items():
                rt_inputs[k] = v
        elif isinstance(X, list):
            if len(self.inputs_) == 1:
                rt_inputs[self.inputs_[0]] = numpy.array(X)
            else:
                for i in range(len(self.inputs_)):  # pylint: disable=C0200
                    rt_inputs[self.inputs_[i]] = [row[i] for row in X]

        for k, v in inputs.items():
            rt_inputs[k] = v

        names = ([self.output_name]
                 if self.output_name else self.onnxrt_.output_names)
        self._check_arrays(rt_inputs)
        doutputs = self.onnxrt_.run(rt_inputs)
        outputs = [doutputs[n] for n in names]

        if self.reshape:
            n = outputs[0].shape[0]
            outputs = [o.reshape((n, -1)) for o in outputs]

        if self.output_name or len(outputs) == 1:
            if isinstance(outputs[0], list):
                return pandas.DataFrame(outputs[0])
            return outputs[0]

        names = self.output_name if self.output_name else [
            o for o in self.onnxrt_.output_names]
        concat = []
        colnames = []
        for k, v in zip(names, outputs):
            if isinstance(v, numpy.ndarray):
                if len(v.shape) == 1:
                    v = v.reshape((-1, 1))
                    colnames.append(k)
                elif len(v.shape) == 2:
                    colnames.extend("%s%d" % (k, i) for i in range(v.shape[1]))
                else:
                    raise RuntimeError(  # pragma: no cover
                        f"Unexpected shape for results {k!r}: {v.shape!r}.")
            if isinstance(v, list):
                if len(v) == 0:
                    raise RuntimeError(  # pragma: no cover
                        f"Output {k!r} is empty.")
                if not isinstance(v[0], dict):
                    raise RuntimeError(  # pragma: no cover
                        f"Unexpected type for output {k!r} - value={v[0]!r}.")
                df = pandas.DataFrame(v)
                cols = list(sorted(df.columns))
                v = df[cols].copy().values
                colnames.extend("%s%d" % (k, i) for i in range(v.shape[1]))
            concat.append(v)
        res = numpy.hstack(concat)
        return pandas.DataFrame(res, columns=colnames)

    def fit_transform(self, X, y=None, **inputs):
        """
        Loads the *ONNX* model and runs the predictions.

        :param X: iterable, data to process
            (or first input if several expected)
        :param y: unused
        :param inputs: :epkg:`ONNX` graph support multiple inputs,
            each column of a dataframe is converted into as many inputs if
            *X* is a dataframe, otherwise, *X* is considered as the first input
            and *inputs* can be used to specify the other ones
        :return: :epkg:`DataFrame`
        """
        return self.fit(X, y=y, **inputs).transform(X, y)

    @staticmethod
    def enumerate_create(onnx_bytes, output_names=None, enforce_float32=True):
        """
        Creates multiple *OnnxTransformer*,
        one for each requested intermediate node.

        onnx_bytes : bytes
        output_names: string
            requested output names or None to request all and
            have method *transform* to store all of them in a dataframe
        enforce_float32 : boolean
            :epkg:`onnxruntime` only supports *float32*,
            :epkg:`scikit-learn` usually uses double floats, this parameter
            ensures that every array of double floats is converted into
            single floats
        :return: iterator on OnnxTransformer *('output name', OnnxTransformer)*
        """
        selected = None if output_names is None else set(output_names)
        model = onnx.load(BytesIO(onnx_bytes))
        for out in enumerate_model_node_outputs(model):
            m = select_model_inputs_outputs(model, out)
            if selected is None or out in selected:
                tr = OnnxTransformer(m.SerializeToString(),
                                     enforce_float32=enforce_float32)
                yield out, tr

    def onnx_parser(self):
        """
        Returns a parser for this model.
        """
        def parser(scope=None, inputs=None):
            if scope is None:
                raise RuntimeError(  # pragma: no cover
                    f"scope cannot be None (parser of class {type(self)!r}).")
            if inputs is None:
                raise RuntimeError(  # pragma: no cover
                    f"inputs cannot be None (parser of class {type(self)!r}).")
            if (not hasattr(self, 'onnxrt_') or
                    not hasattr(self.onnxrt_, 'output_names')):
                raise RuntimeError(  # pragma: no cover
                    'OnnxTransformer not fit.')
            if len(inputs) != len(self.inputs_):
                raise RuntimeError(  # pragma: no cover
                    "Mismatch between the number of inputs, expected %r, "
                    "got %r." % (self.inputs_, inputs))
            return self.onnxrt_.output_names
        return parser

    def onnx_shape_calculator(self):
        def shape_calculator(operator):
            from skl2onnx.common.data_types import (  # delayed
                FloatTensorType, DoubleTensorType, Int64TensorType)
            cout = self.onnxrt_.output_names
            if len(operator.outputs) != len(cout):
                raise RuntimeError(  # pragma: no cover
                    "Mismatched number of outputs: {} != {}."
                    "".format(len(operator.outputs), len(cout)))
            for out_op, out in zip(operator.outputs, self.onnxrt_.obj.graph.output):
                var = _var_as_dict(out)
                if var['type']['kind'] != 'tensor':
                    raise NotImplementedError(  # pragma: no cover
                        f"Noy yet implemented for output:\n{out}")
                shape = var['type']['shape']
                if shape[0] == 0:
                    shape = (None,) + tuple(shape[1:])
                elem = var['type']['elem']
                if elem == 'float':
                    out_op.type = FloatTensorType(shape=shape)
                elif elem == 'int64':
                    out_op.type = Int64TensorType(shape=shape)
                elif elem == 'double':
                    out_op.type = DoubleTensorType(shape=shape)
                else:
                    raise NotImplementedError(  # pragma: no cover
                        f"Not yet implemented for elem_type: {elem!r}")
        return shape_calculator

    def onnx_converter(self):
        """
        Returns a converter for this model.
        If not overloaded, it fetches the converter
        mapped to the first *scikit-learn* parent
        it can find.
        """
        def converter(scope, operator, container, onnx_model=None):
            op = operator.raw_operator
            onx = onnx_model or op.onnxrt_.obj
            add_onnx_graph(scope, operator, container, onx)

        return converter

    @property
    def opsets(self):
        """
        Returns the opsets as dictionary ``{domain: opset}``.
        """
        if hasattr(self, 'onnxrt_'):
            model = self.onnxrt_.obj
        else:
            model = onnx.load(BytesIO(self.onnx_bytes))
        res = {}
        for oimp in model.opset_import:
            res[oimp.domain] = oimp.version
        return res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from ._op import OpRun


def _apply_momentum(r, t, x, g, v, norm_coefficient, alpha, beta):
    # Add gradient of regularization term.
    g_regularized = norm_coefficient * x + g
    # Coefficient of gradient should be 1 at the first iteration.
    beta_adjusted = beta if t > 0 else 1
    # Update momentum.
    v_new = alpha * v + beta_adjusted * g_regularized
    # Apply SG with momentum update rule.
    x_new = x - r * v_new
    return x_new, v_new


class Momentum(OpRun):

    atts = {'alpha': 0,
            'beta': 0,
            'mode': b'standard',
            'norm_coefficient': 0.}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Momentum.atts,
                       **options)

    def _run(self, *data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(data) == 5:
            return self._run1(*data)
        n = (len(data) - 2) // 3
        xs = []
        vs = []
        for i in range(0, n):
            a, b = self._run1(*data[:2], data[2 + i],
                              data[2 + n + i], data[2 + n * 2 + i])
            xs.append(a)
            vs.append(b)
        return tuple(xs + vs)

    def _run1(self, r, t, x, g, v):  # pylint: disable=W0221
        x_new, v_new = _apply_momentum(
            r, t, x, g, v, self.norm_coefficient, self.alpha, self.beta)
        return x_new, v_new



"""
@file
@brief Action definition.
"""
from .gtypes import MLNumTypeFloat32, MLNumTypeFloat64, MLNumTypeBool
from .gactions import MLActionBinary, MLActionFunctionCall


class MLActionAdd(MLActionBinary):
    """
    Addition
    """

    def __init__(self, act1, act2):
        """
        @param  act1    first element
        @param  act2    second element
        """
        MLActionBinary.__init__(self, act1, act2, "+")
        if type(act1.output) != type(act2.output):
            raise TypeError(  # pragma: no cover
                f"Not the same input type {type(act1.output)} != {type(act2.output)}")

    def execute(self, **kwargs):
        MLActionBinary.execute(self, **kwargs)
        res = self.ChildrenResults
        return self.output.validate(res[0] + res[1])


class MLActionSign(MLActionFunctionCall):
    """
    Sign of an expression: 1=positive, 0=negative.
    """

    def __init__(self, act1):
        """
        @param  act1    first element
        """
        MLActionFunctionCall.__init__(self, "sign", act1.output, act1)
        if not isinstance(act1.output, (MLNumTypeFloat32, MLNumTypeFloat64)):
            raise TypeError(  # pragma: no cover
                f"The input action must produce float32 or float64 not '{type(act1.output)}'")

    def execute(self, **kwargs):
        MLActionFunctionCall.execute(self, **kwargs)
        res = self.ChildrenResults
        return self.output.validate(self.output.softcast(1 if res[0] >= 0 else 0))


class MLActionTestInf(MLActionBinary):
    """
    Operator ``<``.
    """

    def __init__(self, act1, act2):
        """
        @param  act1    first element
        @param  act2    second element
        """
        MLActionBinary.__init__(self, act1, act2, "<=")
        if type(act1.output) != type(act2.output):
            raise TypeError(  # pragma: no cover
                f"Not the same input type {type(act1.output)} != {type(act2.output)}")
        self.output = MLNumTypeBool()

    def execute(self, **kwargs):
        MLActionBinary.execute(self, **kwargs)
        res = self.ChildrenResults
        return self.output.validate(self.output.softcast(res[0] <= res[1]))


class MLActionTestEqual(MLActionBinary):
    """
    Operator ``==``.
    """

    def __init__(self, act1, act2):
        """
        @param  act1    first element
        @param  act2    second element
        """
        MLActionBinary.__init__(self, act1, act2, "==")
        if type(act1.output) != type(act2.output):
            raise TypeError(  # pragma: no cover
                f"Not the same input type {type(act1.output)} != {type(act2.output)}")
        self.output = MLNumTypeBool()

    def execute(self, **kwargs):
        MLActionBinary.execute(self, **kwargs)
        res = self.ChildrenResults
        return self.output.validate(self.output.softcast(res[0] == res[1]))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun
from ._new_ops import OperatorSchema


class BroadcastGradientArgs(OpRun):

    atts = {}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)

    def _find_custom_operator_schema(self, op_name):
        if op_name == "BroadcastGradientArgs":
            return BroadcastGradientArgsSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def _run(self, a_shape, b_shape, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221

        A_dims = a_shape
        B_dims = b_shape
        a_size = len(a_shape)
        b_size = len(b_shape)

        ndim = max(a_size, b_size)

        i = a_size - 1
        j = b_size - 1
        k = ndim - 1

        a_axes = []
        b_axes = []

        while i >= 0 and j >= 0:
            A_dim = A_dims[i]
            B_dim = B_dims[j]

            if A_dim != B_dim:
                if A_dim == 1:
                    a_axes.append(k)
                elif B_dim == 1:
                    b_axes.append(k)
                else:
                    a = A_dims[:a_size]
                    b = B_dims[:b_size]
                    raise RuntimeError(
                        "Broadcast is not possible between inputs of "
                        "shapes: %r and %r." % (a, b))
            i -= 1
            j -= 1
            k -= 1

        if i < 0:
            while k >= 0:
                a_axes.append(k)
                k -= 1
        else:
            while k >= 0:
                b_axes.append(k)
                k -= 1

        return (numpy.array(a_axes, dtype=numpy.int64),
                numpy.array(b_axes, dtype=numpy.int64))


class BroadcastGradientArgsSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl BroadcastGradientArgs.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'BroadcastGradientArgs')
        self.attributes = BroadcastGradientArgs.atts



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _global_average_pool(x):
    spatial_shape = numpy.ndim(x) - 2
    y = numpy.average(
        x, axis=tuple(range(spatial_shape, spatial_shape + 2)))
    for _ in range(spatial_shape):
        y = numpy.expand_dims(y, -1)
    return y


def _global_max_pool(x):
    spatial_shape = numpy.ndim(x) - 2
    y = x.max(axis=tuple(range(spatial_shape, spatial_shape + 2)))
    for _ in range(spatial_shape):
        y = numpy.expand_dims(y, -1)
    return y


class GlobalAveragePool(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        res = _global_average_pool(x)
        return (res, )


class GlobalMaxPool(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        res = _global_max_pool(x)
        return (res, )



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *onnx_conv*. Importing this file
means importing :epkg:`sklearn-onnx`.
"""
import onnx
from .register import register_converters, register_scorers
from .register_rewritten_converters import register_rewritten_operators
from .convert import (
    to_onnx, guess_schema_from_data, guess_schema_from_model,
    get_inputs_from_data)



"""
@file
@brief Shortcuts to *onnx_ops*.
"""

from .onnx_complex import OnnxComplexAbs_1, OnnxComplexAbs
from .onnx_fft import (
    OnnxFFT_1, OnnxFFT,
    OnnxRFFT_1, OnnxRFFT,
    OnnxFFT2D_1, OnnxFFT2D)
from .onnx_gradient_op import (
    OnnxYieldOp, OnnxYieldOp_1,
    OnnxBroadcastGradientArgs, OnnxBroadcastGradientArgs_1,
    OnnxFusedMatMul, OnnxFusedMatMul_1,
    OnnxSoftmaxGrad, OnnxSoftmaxGrad_13)
from .onnx_tokenizer import OnnxTokenizer_1, OnnxTokenizer



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE
from ._op import OpRun


class _CommonRandom(OpRun):
    """
    Common methods to all random operators.
    """

    def __init__(self, *args, **kwargs):
        OpRun.__init__(self, *args, **kwargs)

    def _dtype(self, *data, dtype_first=False):
        if dtype_first:
            if self.dtype != 0:
                return self.numpy_type
            if len(data) > 0:
                return data[0].dtype
            raise RuntimeError(  # pragma: no cover
                "dtype cannot be None for operator %s, "
                "self.numpy_type=%r, len(data)=%r."
                "" % (self.__class__.__name__,
                      self.numpy_type, len(data)))
        res = None
        if len(data) == 0:
            res = self.numpy_type
        elif self.numpy_type is not None:
            res = self.numpy_type
        elif hasattr(data[0], 'dtype'):
            res = data[0].dtype
        if res is None:
            raise RuntimeError(  # pragma: no cover
                "dtype cannot be None for operator %s, "
                "self.numpy_type=%r, type(data[0])=%r."
                "" % (self.__class__.__name__,
                      self.numpy_type, type(data[0])))
        return res

    def _get_state(self, seed):
        if numpy.isnan(self.seed):
            state = numpy.random.RandomState()
        else:
            state = numpy.random.RandomState(seed=self.seed)
        return state


class Bernoulli(_CommonRandom):

    atts = {'dtype': 0,
            'seed': numpy.nan}

    def __init__(self, onnx_node, desc=None, **options):
        _CommonRandom.__init__(self, onnx_node, desc=desc,
                               expected_attributes=Bernoulli.atts,
                               **options)
        self.numpy_type = (
            TENSOR_TYPE_TO_NP_TYPE[self.dtype] if self.dtype > 0
            else None)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        dtype = self._dtype(x, dtype_first=True)
        state = self._get_state(self.seed)
        res = state.binomial(1, p=x).astype(dtype)
        return (res.astype(dtype), )

    def to_python(self, inputs):
        lines = [
            'numpy_dtype = TENSOR_TYPE_TO_NP_TYPE[dtype]',
            'state = numpy.random.RandomState(seed=seed)',
            f'return state.binomial(1, {inputs[0]}).astype(numpy_dtype)']
        return ("import numpy\nfrom numpy import nan\n"
                "from onnx.mapping import TENSOR_TYPE_TO_NP_TYPE",
                "\n".join(lines))


class RandomUniform(_CommonRandom):

    atts = {'dtype': 1,
            'low': 0.,
            'high': 1.,
            'seed': numpy.nan,
            'shape': []}

    def __init__(self, onnx_node, desc=None, **options):
        _CommonRandom.__init__(self, onnx_node, desc=desc,
                               expected_attributes=RandomUniform.atts,
                               **options)
        if len(self.shape) == 0:
            raise ValueError(  # pragma: no cover
                f"shape cannot be empty for operator {self.__class__.__name__}.")
        self.numpy_type = TENSOR_TYPE_TO_NP_TYPE[self.dtype]

    def _run(self, *args, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(args) != 0:
            raise RuntimeError(  # pragma: no cover
                f"Operator {self.__class__.__name__} cannot have inputs.")
        dtype = self._dtype(*args)
        state = self._get_state(self.seed)
        res = state.rand(*self.shape).astype(dtype)
        res *= (self.high - self.low)
        res += self.low
        return (res.astype(dtype), )

    def to_python(self, inputs):
        lines = [
            'numpy_dtype = TENSOR_TYPE_TO_NP_TYPE[dtype]',
            'state = numpy.random.RandomState(seed=seed)',
            'return (state.rand(*%r).astype(numpy.%s) * (%f - %f)) + %f' % (
                list(self.shape), self.numpy_type, self.high, self.low, self.low)]
        return ("import numpy\nfrom onnx.mapping import TENSOR_TYPE_TO_NP_TYPE",
                "\n".join(lines))


class RandomUniformLike(_CommonRandom):

    atts = {'low': 0.,
            'high': 1.,
            'seed': numpy.nan,
            'dtype': 0}

    def __init__(self, onnx_node, desc=None, **options):
        _CommonRandom.__init__(self, onnx_node, desc=desc,
                               expected_attributes=RandomUniformLike.atts,
                               **options)
        self.numpy_type = (
            None if self.dtype == 0 else TENSOR_TYPE_TO_NP_TYPE[self.dtype])

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        dtype = self._dtype(x)
        state = self._get_state(self.seed)
        res = state.rand(*x.shape).astype(dtype)
        res *= (self.high - self.low)
        res += self.low
        return (res.astype(dtype), )

    def to_python(self, inputs):
        if len(inputs) > 0 and hasattr(inputs[0], 'dtype'):
            dtype = inputs[0].dtype
            shape = inputs[0].shape
        else:
            dtype = self.numpy_type or numpy.float32
            shape = (1, )
        lines = [
            'numpy_dtype = TENSOR_TYPE_TO_NP_TYPE[dtype]',
            'state = numpy.random.RandomState(seed=seed)',
            'return (state.rand(*%r).astype(numpy.%s) * (%f - %f)) + %f' % (
                shape, dtype, self.high, self.low, self.low)]
        return ("import numpy\nfrom onnx.mapping import TENSOR_TYPE_TO_NP_TYPE",
                "\n".join(lines))


class RandomNormal(_CommonRandom):

    atts = {'dtype': 1,
            'mean': 0.,
            'scale': 1.,
            'seed': numpy.nan,
            'shape': []}

    def __init__(self, onnx_node, desc=None, **options):
        _CommonRandom.__init__(self, onnx_node, desc=desc,
                               expected_attributes=RandomNormal.atts,
                               **options)
        if len(self.shape) == 0:
            raise ValueError(  # pragma: no cover
                f"shape cannot be empty for operator {self.__class__.__name__}.")
        self.numpy_type = TENSOR_TYPE_TO_NP_TYPE[self.dtype]

    def _run(self, *args, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if len(args) != 0:
            raise RuntimeError(  # pragma: no cover
                f"Operator {self.__class__.__name__} cannot have inputs.")
        state = self._get_state(self.seed)
        res = state.randn(*self.shape).astype(self.numpy_type)
        res *= self.scale
        res += self.mean
        return (res.astype(self.numpy_type), )

    def to_python(self, inputs):
        lines = [
            'numpy_dtype = TENSOR_TYPE_TO_NP_TYPE[dtype]',
            'state = numpy.random.RandomState(seed=seed)',
            'return (state.randn(*%r).astype(numpy.%s) * %f) + %f' % (
                list(self.shape), self.numpy_type, self.scale, self.mean)]
        return ("import numpy\nfrom onnx.mapping import TENSOR_TYPE_TO_NP_TYPE",
                "\n".join(lines))


class RandomNormalLike(_CommonRandom):

    atts = {'dtype': 0,
            'mean': 0.,
            'scale': 1.,
            'seed': numpy.nan}

    def __init__(self, onnx_node, desc=None, **options):
        _CommonRandom.__init__(self, onnx_node, desc=desc,
                               expected_attributes=RandomNormalLike.atts,
                               **options)
        self.numpy_type = (
            None if self.dtype == 0 else TENSOR_TYPE_TO_NP_TYPE[self.dtype])

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        dtype = self._dtype(x)
        state = self._get_state(self.seed)
        res = state.randn(*x.shape).astype(dtype)
        res *= self.scale
        res += self.mean
        return (res.astype(dtype), )

    def to_python(self, inputs):
        if len(inputs) > 0 and hasattr(inputs[0], 'dtype'):
            dtype = inputs[0].dtype
            shape = inputs[0].shape
        else:
            dtype = self.numpy_type or numpy.float32
            shape = (1, )
        lines = [
            'numpy_dtype = TENSOR_TYPE_TO_NP_TYPE[dtype]',
            'state = numpy.random.RandomState(seed=seed)',
            'return (state.randn(%r).astype(numpy.%s) * %f) + %f' % (
                shape, dtype, self.scale, self.mean)]
        return ("import numpy\nfrom onnx.mapping import TENSOR_TYPE_TO_NP_TYPE",
                "\n".join(lines))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _one_hot(indices, depth, axis=-1, dtype=numpy.float32):
    values = numpy.asarray(indices)
    rank = len(values.shape)
    depth_range = numpy.arange(depth)
    if axis < 0:
        axis += (rank + 1)
    ls = values.shape[0:axis]
    rs = values.shape[axis:rank]
    new_shape = (1,) * len(ls) + depth_range.shape + (1,) * len(rs)
    targets = numpy.reshape(depth_range, new_shape)
    values = numpy.reshape(numpy.mod(values, depth), ls + (1,) + rs)
    return numpy.asarray(targets == values, dtype=dtype)


class OneHot(OpRun):

    atts = {'axis': -1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=OneHot.atts,
                       **options)

    def _run(self, indices, depth, values, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        off_value, on_value = values
        y = _one_hot(indices, depth, dtype=values.dtype)
        y = y * (on_value - off_value) + off_value
        return (y, )



"""
@file
@brief Command line about validation of prediction runtime.
"""
import os
from collections import OrderedDict
import json
import numpy
from onnx import TensorProto
from pandas import DataFrame
from .. import OnnxInference
from ..ops_whole.session import OnnxWholeSession


def _random_input(typ, shape, batch):
    if typ in ('tensor(double)', TensorProto.DOUBLE):  # pylint: disable=E1101
        dtype = numpy.float64
    elif typ in ('tensor(float)', TensorProto.FLOAT):  # pylint: disable=E1101
        dtype = numpy.float32
    else:
        raise NotImplementedError(
            f"Unable to guess dtype from {typ!r}.")

    if len(shape) <= 1:
        new_shape = shape
    elif shape[0] in (None, 0):
        new_shape = tuple([batch] + list(shape[1:]))
    else:
        new_shape = shape
    return numpy.random.randn(*new_shape).astype(dtype)


def random_feed(inputs, batch=10, empty_dimension=1):
    """
    Creates a dictionary of random inputs.

    :param batch: dimension to use as batch dimension if unknown
    :param empty_dimension: if a dimension is null, replaces it by this value
    :return: dictionary
    """
    res = OrderedDict()
    for inp in inputs:
        name = inp.name
        if hasattr(inp.type, 'tensor_type'):
            typ = inp.type.tensor_type.elem_type
            shape = tuple(getattr(d, 'dim_value', batch)
                          for d in inp.type.tensor_type.shape.dim)
            shape = (shape[0], ) + tuple(
                b if b > 0 else empty_dimension for b in shape[1:])
        else:
            typ = inp.type
            shape = inp.shape
        res[name] = _random_input(typ, shape, batch)
    return res


def latency(model, law='normal', size=1, number=10, repeat=10, max_time=0,
            runtime="onnxruntime", device='cpu', profiling=None):
    """
    Measures the latency of a model (python API).

    :param model: ONNX graph
    :param law: random law used to generate fake inputs
    :param size: batch size, it replaces the first dimension
        of every input if it is left unknown
    :param number: number of calls to measure
    :param repeat: number of times to repeat the experiment
    :param max_time: if it is > 0, it runs as many time during
        that period of time
    :param runtime: available runtime
    :param device: device, `cpu`, `cuda:0`
    :param profiling: if True, profile the execution of every
        node, if can be sorted by name or type,
        the value for this parameter should e in `(None, 'name', 'type')`,
    :return: dictionary or a tuple (dictionary, dataframe)
        if the profiling is enable

    .. cmdref::
        :title: Measures model latency
        :cmd: -m mlprodict latency --help
        :lid: l-cmd-latency

        The command generates random inputs and call many times the
        model on these inputs. It returns the processing time for one
        iteration.

        Example::

            python -m mlprodict latency --model "model.onnx"
    """
    from cpyquickhelper.numbers import measure_time  # delayed import

    if isinstance(model, str) and not os.path.exists(model):
        raise FileNotFoundError(  # pragma: no cover
            f"Unable to find model {model!r}.")
    if profiling not in (None, '', 'name', 'type'):
        raise ValueError(
            f"Unexpected value for profiling: {profiling!r}.")
    size = int(size)
    number = int(number)
    repeat = int(repeat)
    if max_time in (None, 0, ""):
        max_time = None
    else:
        max_time = float(max_time)
        if max_time <= 0:
            max_time = None

    if law != "normal":
        raise ValueError(
            f"Only law='normal' is supported, not {law!r}.")

    if device in ('cpu', 'CPUExecutionProviders'):
        providers = ['CPUExecutionProviders']
    elif device in ('cuda:0', 'CUDAExecutionProviders'):
        if runtime != 'onnxruntime':
            raise NotImplementedError(  # pragma: no cover
                "Only runtime 'onnxruntime' supports this device or provider "
                "%r." % device)
        providers = ['CUDAExecutionProviders']
    elif ',' in device:
        from onnxruntime import get_all_providers  # delayed import
        if runtime != 'onnxruntime':
            raise NotImplementedError(  # pragma: no cover
                "Only runtime 'onnxruntime' supports this device or provider "
                "%r." % device)
        providers = device.split(',')
        allp = set(get_all_providers())
        for p in providers:
            if p not in allp:
                raise ValueError(
                    f"One device or provider {p!r} is not supported among {allp!r}.")
    else:
        raise ValueError(  # pragma no cover
            f"Device {device!r} not supported.")

    if runtime in ("onnxruntime", "onnxruntime-cuda"):
        from onnxruntime import InferenceSession, SessionOptions  # delayed import
        providers = ['CPUExecutionProvider']
        if runtime == "onnxruntime-cuda":
            providers = ['CUDAExecutionProvider'] + providers
        if profiling in ('name', 'type'):
            so = SessionOptions()
            so.enable_profiling = True
            sess = InferenceSession(
                model, sess_options=so, providers=providers)
        else:
            sess = InferenceSession(model, providers=providers)
        fct = lambda feeds: sess.run(None, feeds)
        inputs = sess.get_inputs()
    else:
        if profiling in ('name', 'type'):
            runtime_options = {"enable_profiling": True}
            if runtime != 'onnxruntime1':
                raise NotImplementedError(  # pragma: no cover
                    f"Profiling is not implemented for runtime={runtime!r}.")
        else:
            runtime_options = None
        oinf = OnnxInference(model, runtime=runtime,
                             runtime_options=runtime_options)
        fct = lambda feeds: oinf.run(feeds)
        inputs = oinf.obj.graph.input

    feeds = random_feed(inputs, size)
    res = measure_time(
        lambda: fct(feeds), number=number, repeat=repeat, context={},
        max_time=max_time, div_by_number=True)
    for k, v in feeds.items():
        res[f"shape({k})"] = "x".join(map(str, v.shape))
    if profiling in ('name', 'type'):
        if runtime == 'onnxruntime':
            profile_name = sess.end_profiling()
            with open(profile_name, 'r', encoding='utf-8') as f:
                js = json.load(f)
            js = OnnxWholeSession.process_profiling(js)
            df = DataFrame(js)
        else:
            df = oinf.get_profiling(as_df=True)
        if profiling == 'name':
            gr = df[['dur', "args_op_name", "name"]].groupby(
                ["args_op_name", "name"]).sum().sort_values('dur')
        else:
            gr = df[['dur', "args_op_name"]].groupby(
                "args_op_name").sum().sort_values('dur')
        return res, gr

    return res



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
from collections import OrderedDict
import numpy
from onnx.defs import onnx_opset_version
from ._op_helper import _get_typed_class_attribute
from ._op import OpRunClassifierProb, RuntimeTypeError
from ._op_classifier_string import _ClassifierCommon
from ._new_ops import OperatorSchema
from .op_tree_ensemble_classifier_ import (  # pylint: disable=E0611,E0401
    RuntimeTreeEnsembleClassifierDouble,
    RuntimeTreeEnsembleClassifierFloat)
from .op_tree_ensemble_classifier_p_ import (  # pylint: disable=E0611,E0401
    RuntimeTreeEnsembleClassifierPFloat,
    RuntimeTreeEnsembleClassifierPDouble)


class TreeEnsembleClassifierCommon(OpRunClassifierProb, _ClassifierCommon):

    def __init__(self, dtype, onnx_node, desc=None,
                 expected_attributes=None,
                 runtime_version=3, **options):
        OpRunClassifierProb.__init__(
            self, onnx_node, desc=desc,
            expected_attributes=expected_attributes, **options)
        self._init(dtype=dtype, version=runtime_version)

    def _get_typed_attributes(self, k):
        return _get_typed_class_attribute(self, k, self.__class__.atts)

    def _find_custom_operator_schema(self, op_name):
        """
        Finds a custom operator defined by this runtime.
        """
        if op_name == "TreeEnsembleClassifierDouble":
            return TreeEnsembleClassifierDoubleSchema()
        raise RuntimeError(  # pragma: no cover
            f"Unable to find a schema for operator '{op_name}'.")

    def _init(self, dtype, version):
        self._post_process_label_attributes()

        atts = []
        for k in self.__class__.atts:
            v = self._get_typed_attributes(k)
            if k.endswith('_as_tensor'):
                if (v is not None and isinstance(v, numpy.ndarray) and
                        v.size > 0):
                    # replacements
                    atts[-1] = v
                    if dtype is None:
                        dtype = v.dtype
                continue
            atts.append(v)

        if dtype is None:
            dtype = numpy.float32

        if dtype == numpy.float32:
            if version == 0:
                self.rt_ = RuntimeTreeEnsembleClassifierFloat()
            elif version == 1:
                self.rt_ = RuntimeTreeEnsembleClassifierPFloat(
                    60, 20, False, False)
            elif version == 2:
                self.rt_ = RuntimeTreeEnsembleClassifierPFloat(
                    60, 20, True, False)
            elif version == 3:
                self.rt_ = RuntimeTreeEnsembleClassifierPFloat(
                    60, 20, True, True)
            else:
                raise ValueError(f"Unknown version '{version}'.")
        elif dtype == numpy.float64:
            if version == 0:
                self.rt_ = RuntimeTreeEnsembleClassifierDouble()
            elif version == 1:
                self.rt_ = RuntimeTreeEnsembleClassifierPDouble(
                    60, 20, False, False)
            elif version == 2:
                self.rt_ = RuntimeTreeEnsembleClassifierPDouble(
                    60, 20, True, False)
            elif version == 3:
                self.rt_ = RuntimeTreeEnsembleClassifierPDouble(
                    60, 20, True, True)
            else:
                raise ValueError(  # pragma: no cover
                    f"Unknown version '{version}'.")
        else:
            raise RuntimeTypeError(  # pragma: no cover
                f"Unsupported dtype={dtype}.")
        self.rt_.init(*atts)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        """
        This is a C++ implementation coming from
        :epkg:`onnxruntime`.
        `tree_ensemble_classifier.cc
        <https://github.com/microsoft/onnxruntime/blob/master/
        onnxruntime/core/providers/cpu/ml/tree_ensemble_classifier.cc>`_.
        See class :class:`RuntimeTreeEnsembleClassifier
        <mlprodict.onnxrt.ops_cpu.op_tree_ensemble_classifier_.RuntimeTreeEnsembleClassifier>`.
        """
        label, scores = self.rt_.compute(x)
        if scores.shape[0] != label.shape[0]:
            scores = scores.reshape(label.shape[0],
                                    scores.shape[0] // label.shape[0])
        return self._post_process_predicted_label(label, scores)


class TreeEnsembleClassifier_1(TreeEnsembleClassifierCommon):

    atts = OrderedDict([
        ('base_values', numpy.empty(0, dtype=numpy.float32)),
        ('class_ids', numpy.empty(0, dtype=numpy.int64)),
        ('class_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('class_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('class_weights', numpy.empty(0, dtype=numpy.float32)),
        ('classlabels_int64s', numpy.empty(0, dtype=numpy.int64)),
        ('classlabels_strings', []),
        ('nodes_falsenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_featureids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_hitrates', numpy.empty(0, dtype=numpy.float32)),
        ('nodes_missing_value_tracks_true', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_modes', []),
        ('nodes_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_truenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_values', numpy.empty(0, dtype=numpy.float32)),
        ('post_transform', b'NONE')
    ])

    def __init__(self, onnx_node, desc=None, **options):
        TreeEnsembleClassifierCommon.__init__(
            self, numpy.float32, onnx_node, desc=desc,
            expected_attributes=TreeEnsembleClassifier_1.atts, **options)


class TreeEnsembleClassifier_3(TreeEnsembleClassifierCommon):

    atts = OrderedDict([
        ('base_values', numpy.empty(0, dtype=numpy.float32)),
        ('base_values_as_tensor', []),
        ('class_ids', numpy.empty(0, dtype=numpy.int64)),
        ('class_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('class_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('class_weights', numpy.empty(0, dtype=numpy.float32)),
        ('class_weights_as_tensor', []),
        ('classlabels_int64s', numpy.empty(0, dtype=numpy.int64)),
        ('classlabels_strings', []),
        ('nodes_falsenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_featureids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_hitrates', numpy.empty(0, dtype=numpy.float32)),
        ('nodes_hitrates_as_tensor', []),
        ('nodes_missing_value_tracks_true', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_modes', []),
        ('nodes_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_truenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_values', numpy.empty(0, dtype=numpy.float32)),
        ('nodes_values_as_tensor', []),
        ('post_transform', b'NONE')
    ])

    def __init__(self, onnx_node, desc=None, **options):
        TreeEnsembleClassifierCommon.__init__(
            self, None, onnx_node, desc=desc,
            expected_attributes=TreeEnsembleClassifier_3.atts, **options)


class TreeEnsembleClassifierDouble(TreeEnsembleClassifierCommon):

    atts = OrderedDict([
        ('base_values', numpy.empty(0, dtype=numpy.float64)),
        ('class_ids', numpy.empty(0, dtype=numpy.int64)),
        ('class_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('class_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('class_weights', numpy.empty(0, dtype=numpy.float64)),
        ('classlabels_int64s', numpy.empty(0, dtype=numpy.int64)),
        ('classlabels_strings', []),
        ('nodes_falsenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_featureids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_hitrates', numpy.empty(0, dtype=numpy.float64)),
        ('nodes_missing_value_tracks_true', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_modes', []),
        ('nodes_nodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_treeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_truenodeids', numpy.empty(0, dtype=numpy.int64)),
        ('nodes_values', numpy.empty(0, dtype=numpy.float64)),
        ('post_transform', b'NONE')
    ])

    def __init__(self, onnx_node, desc=None, **options):
        TreeEnsembleClassifierCommon.__init__(
            self, numpy.float64, onnx_node, desc=desc,
            expected_attributes=TreeEnsembleClassifierDouble.atts, **options)


class TreeEnsembleClassifierDoubleSchema(OperatorSchema):
    """
    Defines a schema for operators added in this package
    such as @see cl TreeEnsembleClassifierDouble.
    """

    def __init__(self):
        OperatorSchema.__init__(self, 'TreeEnsembleClassifierDouble')
        self.attributes = TreeEnsembleClassifierDouble.atts


if onnx_opset_version() >= 16:
    TreeEnsembleClassifier = TreeEnsembleClassifier_3
else:
    TreeEnsembleClassifier = TreeEnsembleClassifier_1



# -*- coding: utf-8 -*-
"""
@file
@brief Helpers to identify an interpreter.
"""
import keyword
import re
from .g_sklearn_linear_model import sklearn_logistic_regression, sklearn_linear_regression
from .g_sklearn_preprocessing import sklearn_standard_scaler
from .g_sklearn_tree import sklearn_decision_tree_regressor


def __pep8():  # pragma: no cover
    assert sklearn_decision_tree_regressor
    assert sklearn_linear_regression
    assert sklearn_logistic_regression
    assert sklearn_standard_scaler


def change_style(name):
    """
    Switches from *AaBb* into *aa_bb*.

    @param      name    name to convert
    @return             converted name
    """
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    s2 = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()
    return s2 if not keyword.iskeyword(s2) else s2 + "_"


def identify_interpreter(model):
    """
    Identifies the interpreter for a *scikit-learn* model.

    @param      model       model to identify
    @return                 interpreter
    """
    class_name = model.__class__.__name__
    pyname = change_style(class_name)
    skconv = "sklearn_" + pyname
    loc = globals().copy()
    convs = {k: v for k, v in loc.items() if k.startswith("sklearn")}
    if len(convs) == 0:
        raise ValueError(  # pragma: no cover
            "No found interpreters, possibilities=\n{0}".format(
                "\n".join(sorted(loc.keys()))))
    if skconv in convs:
        return convs[skconv]
    raise NotImplementedError(  # pragma: no cover
        "Model class '{0}' is not yet implemented. Available interpreters:\n{1}".format(
            class_name, "\n".join(
                sorted(
                    convs.keys()))))



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


def _specify_int64(indices, inverse_indices, counts):
    return (numpy.array(indices, dtype=numpy.int64),
            numpy.array(inverse_indices, dtype=numpy.int64),
            numpy.array(counts, dtype=numpy.int64))


class Unique(OpRun):

    atts = {'axis': numpy.nan, 'sorted': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       expected_attributes=Unique.atts,
                       **options)

    def _run(self, x, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if numpy.isnan(self.axis):
            y, indices, inverse_indices, counts = numpy.unique(
                x, True, True, True)
        else:
            y, indices, inverse_indices, counts = numpy.unique(
                x, True, True, True, axis=self.axis)
        if len(self.onnx_node.output) == 1:
            return (y, )

        if not self.sorted:
            argsorted_indices = numpy.argsort(indices)
            inverse_indices_map = {
                i: si
                for i, si in zip(
                    argsorted_indices, numpy.arange(len(argsorted_indices)))}
            indices = indices[argsorted_indices]
            y = numpy.take(x, indices, axis=0)
            inverse_indices = numpy.asarray(
                [inverse_indices_map[i] for i in inverse_indices],
                dtype=numpy.int64)
            counts = counts[argsorted_indices]

        indices, inverse_indices, counts = _specify_int64(
            indices, inverse_indices, counts)
        if len(self.onnx_node.output) == 2:
            return (y, indices)
        if len(self.onnx_node.output) == 3:
            return (y, indices, inverse_indices)
        return (y, indices, inverse_indices, counts)



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRunUnary


class IsInf(OpRunUnary):

    atts = {'detect_negative': 1, 'detect_positive': 1}

    def __init__(self, onnx_node, desc=None, **options):
        OpRunUnary.__init__(self, onnx_node, desc=desc,
                            expected_attributes=IsInf.atts,
                            **options)

    def _run(self, data, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        if self.detect_negative:
            if self.detect_positive:
                return (numpy.isinf(data), )
            return (numpy.isneginf(data), )
        if self.detect_positive:
            return (numpy.isposinf(data), )
        res = numpy.full(data.shape, dtype=numpy.bool_, fill_value=False)
        return (res, )

    def to_python(self, inputs):
        return self._to_python_numpy(inputs, 'isnan')



"""
@file
@brief Implements decorators to extend the API.
"""


class AutoAction:
    """
    Extends the API to automatically look for exporters.
    """

    def _reset_cache(self):
        """
        A same node may appear at different places in the graph.
        It means the output is used twice. However, we don't want to
        include the code to generate that same output twice. We cache it
        and keep some information about it.
        """
        self._cache = None
        for child in self.children:  # pylint: disable=E1101
            child._reset_cache()

    def export(self, lang="json", hook=None, result_name=None):
        """
        Exports into any format.
        The method is looking for one method call
        '_export_<lang>' and calls it if found.

        @param      lang            language
        @param      hook            tweaking parameters
        @param      result_name     the name of the result decided by the parent of this node
        @return                     depends on the language
        """
        self._reset_cache()
        name = f"_export_{lang}"
        if hasattr(self, name):
            try:
                return getattr(self, name)(hook=hook, result_name=result_name)
            except TypeError as e:  # pragma: no cover
                raise TypeError(
                    f"Signature of '{name}' is wrong for type '{type(self)}'") from e
        else:
            raise NotImplementedError(  # pragma: no cover
                "No conversion is implemented for lang='{0}' and type='{1}'".format(
                    lang, type(self)))

    @staticmethod
    def cache(func):
        """Caches the result of a function."""

        def func_wrapper(self, hook=None, result_name=None):
            """Wrapper to cache the result of a function."""
            if self._cache is not None:
                c = self._cache.copy()
                c['cache'] = True
                return c
            else:
                ret = func(self, hook=hook, result_name=result_name)
                if not isinstance(ret, dict):
                    raise TypeError(  # pragma: no cover
                        "A dictionary was expected not '{0}'.\nIssue with class '{1}'"
                        "".format(
                            type(ret), type(self)))
                self._cache = ret
                ret = ret.copy()
                ret['cache'] = False
                return ret
        return func_wrapper


class AutoType:
    """
    Extends the API to automatically look for exporters.
    """

    def format_value(self, value, lang="json", hook=None):
        """
        Exports into any format.
        The method is looking for one method call
        '_export_<lang>' and calls it if found.

        @param      value           value to format
        @param      lang            language
        @param      hook            tweaking parameters
        @return                     depends on the language
        """
        name = f"_format_value_{lang}"
        if hasattr(self, name):
            try:
                return getattr(self, name)(value, hook=hook)
            except TypeError as e:
                raise TypeError(
                    f"Singature of '{name}' is wrong for type '{type(self)}'") from e
        else:
            raise NotImplementedError(
                "No formatting is implemented for lang='{0}' and type='{1}'".format(
                    lang, type(self)))



"""
@file
@brief Helpers to run examples created with :epkg:`sklearn-onnx`.
"""
from onnx import helper, TensorProto, ValueInfoProto, TypeProto


def get_tensor_shape(obj):
    """
    Returns the shape if that makes sense for this object.
    """
    if isinstance(obj, ValueInfoProto):
        return get_tensor_shape(obj.type)
    elif not isinstance(obj, TypeProto):
        raise TypeError(  # pragma: no cover
            f"Unexpected type {type(obj)!r}.")
    shape = []
    for d in obj.tensor_type.shape.dim:
        v = d.dim_value if d.dim_value > 0 else d.dim_param
        shape.append(v)
    if len(shape) == 0:
        shape = None
    else:
        shape = list(None if s == 0 else s for s in shape)
    return shape


def get_tensor_elem_type(obj):
    """
    Returns the element type if that makes sense for this object.
    """
    if isinstance(obj, ValueInfoProto):
        return get_tensor_elem_type(obj.type)
    elif not isinstance(obj, TypeProto):
        raise TypeError(  # pragma: no cover
            f"Unexpected type {type(obj)!r}.")
    return obj.tensor_type.elem_type


def _copy_inout(inout, scope, new_name):
    shape = get_tensor_shape(inout)
    elem_type = get_tensor_elem_type(inout)
    value_info = helper.make_tensor_value_info(
        new_name, elem_type, shape)
    return value_info


def _clean_variable_name(name, scope):
    return scope.get_unique_variable_name(name)


def _clean_operator_name(name, scope):
    return scope.get_unique_operator_name(name)


def _clean_initializer_name(name, scope):
    return scope.get_unique_variable_name(name)


def add_onnx_graph(scope, operator, container, onx):
    """
    Adds a whole ONNX graph to an existing one following
    :epkg:`skl2onnx` API assuming this ONNX graph implements
    an `operator <http://onnx.ai/sklearn-onnx/api_summary.html?
    highlight=operator#skl2onnx.common._topology.Operator>`_.

    :param scope: scope (to get unique names)
    :param operator: operator
    :param container: container
    :param onx: ONNX graph
    """
    graph = onx.graph
    name_mapping = {}
    node_mapping = {}
    for node in graph.node:
        name = node.name
        if name is not None:
            node_mapping[node.name] = _clean_initializer_name(
                node.name, scope)
        for o in node.input:
            name_mapping[o] = _clean_variable_name(o, scope)
        for o in node.output:
            name_mapping[o] = _clean_variable_name(o, scope)
    for o in graph.initializer:
        name_mapping[o.name] = _clean_operator_name(o.name, scope)

    inputs = [_copy_inout(o, scope, name_mapping[o.name])
              for o in graph.input]
    outputs = [_copy_inout(o, scope, name_mapping[o.name])
               for o in graph.output]

    for inp, to in zip(operator.inputs, inputs):
        n = helper.make_node('Identity', [inp.onnx_name], [to.name],
                             name=_clean_operator_name('Identity', scope))
        container.nodes.append(n)

    for inp, to in zip(outputs, operator.outputs):
        n = helper.make_node('Identity', [inp.name], [to.onnx_name],
                             name=_clean_operator_name('Identity', scope))
        container.nodes.append(n)

    for node in graph.node:
        n = helper.make_node(
            node.op_type,
            [name_mapping[o] for o in node.input],
            [name_mapping[o] for o in node.output],
            name=node_mapping[node.name] if node.name else None,
            domain=node.domain if node.domain else None)
        n.attribute.extend(node.attribute)  # pylint: disable=E1101
        container.nodes.append(n)

    for o in graph.initializer:
        as_str = o.SerializeToString()
        tensor = TensorProto()
        tensor.ParseFromString(as_str)
        tensor.name = name_mapping[o.name]
        container.initializers.append(tensor)

    # opset
    for oimp in onx.opset_import:
        container.node_domain_version_pair_sets.add(
            (oimp.domain, oimp.version))
